import re
import json
import time
from urllib.parse import quote_plus
import hmac
import base64
import hashlib
import queue

import requests

from linora.utils._config import Config
from linora.utils._logger import Logger

__all__ = ['BotDingTalk']


class BotDingTalk():
    """Dingding group custom robot
    
    each robot can send up to 20 messages per minute, 
    supports three message types: text (text), link (link), and markdown.
    """
    def __init__(self, webhook, secret=None, pc_slide=False, fail_notice=False, logger=None, verbose=0):
        """
        Args:
            webhook: Dingding group custom robot webhook address.
            secret: The secret key that needs to be passed in when "Signing" is checked on the robot security settings page.
            pc_slide: The message link opening method, the default is False to open the browser, 
                      when set to True, the sidebar on the PC side is opened.
            fail_notice: message sending failure reminder, the default is False to not remind, 
                         the developer can judge and deal with it according to the returned message sending result.
            logger: Logger object, la.utils.Logger() class.
            verbose: Verbosity mode, 0 (silent), 1 (verbose).
        """
        warn = Logger(name='')
        warn.info(f"la.utils.message.BotDingTalk has been deprecated and will be deleted in version 2.1.0. Please use la.server.message.BotDingTalk")
        self._params.Config()
        self._params.headers = {'Content-Type': 'application/json; charset=utf-8'}
        self._params.queue = queue.Queue(20)
        self._params.webhook = webhook
        self._params.secret = secret
        self._params.pc_slide = str.lower(pc_slide)
        self._params.fail_notice = fail_notice
        self._params.time_start = time.time()
        if self._params.secret is not None and self._params.secret.startswith('SEC'):
            self._update_webhook()
        if logger is None:
            logger = Logger()
        self._params.logger = logger
        self.params.verbose = verbose
            
    def _update_webhook(self):
        timestamp = round(self._params.time_start * 1000)
        string_to_sign = f'{timestamp}\n{self._params.secret}'
        hmac_code = hmac.new(self._params.secret.encode(), string_to_sign.encode(), digestmod=hashlib.sha256).digest()
        
        sign = quote_plus(base64.b64encode(hmac_code))
        if 'timestamp'in self.webhook:
            self._params.webhook = '{}&timestamp={}&sign={}'.format(self._params.webhook[:self._params.webhook.find('&timestamp')], str(timestamp), sign)
        else:
            self._params.webhook = '{}&timestamp={}&sign={}'.format(self._params.webhook, str(timestamp), sign)
            
    def _format_url(self, url):
        return f'dingtalk://dingtalkclient/page/link?url={quote_plus(url)}&pc_slide={self._params.pc_slide}'
    
    def _assert_msg(self, msg):
        assert len(msg.strip())>0, f"`{msg}` msg cannot be empty."

    def send_text(self, text, is_at_all=False, at_mobiles=[], at_dingtalk_ids=[], msg_add_mobile=True):
        """send text message.
        
        Args:
            text: message content.
            is_at_all: true is at all people, false is not at all people.
            at_mobiles: list of at people mobiles.
            at_dingtalk_ids: list of at people dingtalk ids.
            msg_add_mobile: message auto add to mobile.
        """
        self._assert_msg(text)
        data = {"msgtype": "text", "text":{"content": text}, "at": {}}
        if is_at_all:
            data["at"]["isAtAll"] = 'true'
        if at_mobiles:
            at_mobiles = list(map(str, at_mobiles))
            data["at"]["atMobiles"] = at_mobiles
            if msg_add_mobile:
                data["text"]["content"] = text + '\n@' + '@'.join(at_mobiles)
        if at_dingtalk_ids:
            data["at"]["atDingtalkIds"] = list(map(str, at_dingtalk_ids))
        return self._post(data)

    def send_image(self, image_url):
        """send image message.
        
        Args:
            image_url: image url.
        """
        self._assert_msg(image_url)
        data = {"msgtype":"image", "image":{"picURL":image_url}}
        return self._post(data)

    def send_link(self, text, title, message_url, image_url=''):
        """send link message.
        
        Args:
            text: message content, if it's too long, only part of it will be displayed.
            title: message title.
            message_url: link url, click the url to which the message is redirected.
            image_url: image url.
        """
        self._assert_msg(title)
        self._assert_msg(text)
        self._assert_msg(message_url)
        data = {"msgtype": "link",
                "link": {"text":text, "title":title, "picUrl":image_url, "messageUrl":self._format_url(message_url)}}
        return self._post(data)

    def send_markdown(self, text, title, is_at_all=False, at_mobiles=[], at_dingtalk_ids=[], msg_add_mobile=True):
        """send markdown message.
        
        Args:
            text: Message content in markdown format.
            title: The display content revealed by the above-the-fold session.
            is_at_all: true is at all people, false is not at all people.
            at_mobiles: list of at people mobiles.
            at_dingtalk_ids: list of at people dingtalk ids.
            msg_add_mobile: message auto add to mobile.
        """
        self._assert_msg(title)
        self._assert_msg(text)
        text = re.sub(r'(?<!!)\[.*?\]\((.*?)\)', lambda m: m.group(0).replace(m.group(1), self._format_url(m.group(1))), text)
        data = {"msgtype": "markdown", "markdown": {"title": title, "text": text}, "at": {}}
        if is_at_all:
            data["at"]["isAtAll"] = 'true'
        if at_mobiles:
            at_mobiles = list(map(str, at_mobiles))
            data["at"]["atMobiles"] = at_mobiles
            if is_auto_at:
                data["markdown"]["text"] = text + '\n@' + '@'.join(at_mobiles)
        if at_dingtalk_ids:
            data["at"]["atDingtalkIds"] = list(map(str, at_dingtalk_ids))
        return self._post(data)

    def send_action_card(self, title, text, btns, btn_orientation=0, hide_avatar=0):
        """send ActionCard message.
        
        Args:
            title: Display content from the above-the-fold session.
            text: message in markdown format
            btns: button list, when the number of buttons is 1, the overall jump to ActionCard type;
                  when the number of buttons is greater than 1, the independent jump to ActionCard type.
            btn_orientation: 0: Buttons are arranged vertically, 1: Buttons are arranged horizontally (optional).
            hide_avatar: 0: normal sender avatar, 1: hide sender avatar (optional).
        """
        if len(btns) == 1:
            data = {"msgtype": "actionCard",
                    "actionCard": {
                        "title":title,
                        "text":text,
                        "hideAvatar":hide_avatar,
                        "btnOrientation":btn_orientation,
                        "singleTitle":btns[0]["title"],
                        "singleURL":btns[0]["actionURL"]}}
        else:
            data = {"msgtype": "actionCard",
                    "actionCard": {
                        "title":title,
                        "text":text,
                        "hideAvatar":hide_avatar,
                        "btnOrientation":btn_orientation,
                        "btns":btns}}
        return self._post(data)

    def send_feed_card(self, links):
        """send FeedCard message.
        
        Args:
            links: a list of FeedLink isinstance or CardItem isinstance.
        """
        assert isinstance(links, list), '`links` format error, should be list.'
        link_list = []
        for link in links:
            assert isinstance(link, FeedLink) or isinstance(link, CardItem), f'`{link}` format error'
            link = link.get_data()
            link['messageURL'] = link['messageURL']
            link_list.append(link)
        data = {"msgtype": "feedCard", "feedCard": {"links": link_list}}
        return self._post(data)

    def make_card(self, title, url, image_url=None):
        """make FeedCard or ActionCard.
        
        Args:
            title: message name.
            url: URL triggered by click.
            image_url: default make ActionCard, if image_url is not None, make FeedCard.
        """
        self._assert_msg(title)
        self._assert_msg(url)
        if image_url is None:
            return {"title":title, "actionURL":self._format_url(url)}
        else:
            self._assert_msg(image_url)
            return {"title":title, "messageURL":self._format_url(url), "picURL":image_url}
        
    def _post(self, data):
        time_now = time.time()
        
        if time_now - self._params.time_start >= 3600 and self._params.secret is not None and self._params.secret.startswith('SEC'):
            self._params.time_start = time_now
            self._update_webhook()

        self._params.queue.put(time_now)
        if self._params.queue.full():
            elapse_time = time_now - self._params.queue.get()
            if elapse_time < 60:
                sleep_time = int(60 - elapse_time) + 1
                self._log(f'钉钉官方限制机器人每分钟最多发送20条，当前发送频率已达限制条件，休眠 {sleep_time}s')
                time.sleep(sleep_time)

        try:
            post_data = json.dumps(data)
            response = requests.post(self.webhook, headers=self.headers, data=post_data)
        except requests.exceptions.HTTPError as exc:
            self._log(f'message send fail, HTTP error: {exc.response.status_code}, reason: {exc.response.reason}')
            raise
        except requests.exceptions.ConnectionError:
            self._log("message send fail，HTTP connection error!")
            raise
        except requests.exceptions.Timeout:
            self._log("message send fail，Timeout error!")
            raise
        except requests.exceptions.RequestException:
            self._log("message send fail, Request Exception!")
            raise
        else:
            try:
                result = response.json()
            except JSONDecodeError:
                self._log("服务器响应异常，状态码：%s，响应内容：%s" % (response.status_code, response.text))
                return {'errcode': 500, 'errmsg': '服务器响应异常'}
            else:
                self._log('发送结果：%s' % result)
                # 消息发送失败提醒（errcode 不为 0，表示消息发送异常），默认不提醒，开发者可以根据返回的消息发送结果自行判断和处理
                if self.fail_notice and result.get('errcode', True):
                    time_now = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
                    error_data = {
                      "msgtype": "text",
                      "text": {
                        "content": "[注意-自动通知]钉钉机器人消息发送失败，时间：%s，原因：%s，请及时跟进，谢谢!" % (
                          time_now, result['errmsg'] if result.get('errmsg', False) else '未知异常')
                        },
                      "at": {
                        "isAtAll": False
                        }
                      }
                    self._log("消息发送失败，自动通知：%s" % error_data)
                    requests.post(self.webhook, headers=self.headers, data=json.dumps(error_data))
                return result

    def _log(self, msg):
        if self._params.logger.params.log_file!='':
            self._params.logger.write(msg)
        if self._params.verbose:
            self._params.logger.info(msg)



import math

import numpy as np
from PIL import Image, ImageChops

from linora.image._image_rgb import _fill_color

__all__ = ['flip_up_left', 'flip_up_right', 'flip_left_right', 'flip_up_down', 
           'rotate', 'translate', 'offset', 'pad', 'channel_shuffle',
           'perspective', 'affine', 'shear', 'rescale', 'jigsaw'
          ]


def flip_up_left(image, p=1):
    """Randomly flip an image (up to left).
    
    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    return image.transpose(Image.Transpose.TRANSPOSE)


def flip_up_right(image, p=1):
    """Randomly flip an image (up to right).

    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    return image.transpose(Image.Transpose.TRANSVERSE)


def flip_left_right(image, p=1):
    """Randomly flip an image (left to right).

    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    return image.transpose(Image.Transpose.FLIP_LEFT_RIGHT)


def flip_up_down(image, p=1):
    """Randomly flip an image (up to down).

    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    return image.transpose(Image.Transpose.FLIP_TOP_BOTTOM)


def rotate(image, angle, expand=False, center=(0.5,0.5), translate=(0,0), fill_color=None, p=1):
    """Returns a rotated copy of this image. 
    
    This method returns a copy of this image, rotated the given number of degrees counter clockwise around its centre.
    
    Args:
        image: a PIL instance.
        angle: In degrees counter clockwise, angle in [-180, 180].
               if int or float, rotation angle.
               if list or tuple, randomly picked in the interval `[angle[0], angle[1])` value.
        expand: if true, expands the output image to make it large enough to hold the entire rotated image. 
                if false, make the output image the same size as the input image. 
                Note that the expand flag assumes rotation around the center and no translation.
                if value is None, then the function is random.
        center: center of rotation, xaxis and yaxis in [0,1], default is the center of the image.
                if int or float, xaxis=yaxis,
                if 2-tuple, (xaxis, yaxis), 
                if 4-tuple, xaxis in (center[0], center[1]) and yaxis in (center[2], center[3]).
        translate: post-rotate translation, xoffset and yoffset in [-1,1], see la.image.translate method.
                   if int or float, xoffset=yoffset,
                   if 2-tuple, (xoffset, yoffset), 
                   if 4-tuple, xoffset in (translate[0], translate[1]) and  yoffset in (translate[2], translate[3]).
        fill_color: color for area outside, int or str or tuple or la.image.RGBMode, rgb color.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(angle, (list, tuple)):
        angle = np.random.uniform(angle[0], angle[1])
    if expand is None:
        expand = np.random.choice([True, False])
        
    if isinstance(center, (int, float)):
        center = (int(center*image.size[0]), int(center*image.size[1]))
    elif isinstance(center, (list, tuple)):
        if len(center)==2:
            center = (int(center[0]*image.size[0]), int(center[1]*image.size[1]))
        elif len(center)==4:
            center = (int(np.random.uniform(center[0], center[1])*image.size[0]), 
                      int(np.random.uniform(center[2], center[3])*image.size[1]))
        else:
            raise ValueError('`center` value format error.')
    else:
        raise ValueError('`center` value format error.')
    
    if isinstance(translate, (int, float)):
        translate = (int(translate*image.size[0]), int(translate*image.size[1]))
    elif isinstance(translate, (list, tuple)):
        if len(translate)==2:
            translate = (int(translate[0]*image.size[0]), int(translate[1]*image.size[1]))
        elif len(translate)==4:
            translate = (int(np.random.uniform(translate[0], translate[1])*image.size[0]), 
                         int(np.random.uniform(translate[2], translate[3])*image.size[1]))
        else:
            raise ValueError('`translate` value format error.')
    else:
        raise ValueError('`translate` value format error.')
    
    fill_color = _fill_color(image, fill_color)
    return image.rotate(angle, resample=Image.Resampling.NEAREST, expand=expand, 
                        center=center, translate=translate, fillcolor=fill_color)


def translate(image, xoffset=(-0.5,0.5), yoffset=None, fill_color=None, p=1):
    """Returns a translate copy of this image. 
    
    Args:
        image: a PIL instance.
        xoffset: [-1, 1], int or float, width offset.
                 if list or tuple, randomly picked in the interval `[xoffset[0], xoffset[1])`.
        yoffset: [-1, 1], int or float, height offset.
                 if list or tuple, randomly picked in the interval `[yoffset[0], yoffset[1])`.
        fill_color: int or str or tuple or la.image.RGBMode, rgb color.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(xoffset, (list, tuple)):
        xoffset = np.random.uniform(xoffset[0], xoffset[1])
    if yoffset is None:
        yoffset = xoffset
    elif isinstance(yoffset, (list, tuple)):
        yoffset = np.random.uniform(yoffset[0], yoffset[1])
    xoffset = int(image.size[0]*xoffset)
    yoffset = int(image.size[1]*yoffset)
    fill_color = _fill_color(image, fill_color)
    return rotate(image, angle=0, expand=0, center=None, translate=(xoffset, yoffset), fillcolor=fill_color)


def offset(image, xoffset, yoffset=None, p=1):
    """Returns a copy of the image where data has been offset by the given distances.
    
    Data wraps around the edges. If ``yoffset`` is omitted, it is assumed to be equal to ``xoffset``.

    Args:
        image: a PIL instance.
        xoffset: int or list ot tuple, The horizontal distance.
                 if tuple or list, randomly picked in the interval `[xoffset[0], xoffset[1])`.
        yoffset: int or list ot tuple, The vertical distance. 
                 If omitted, both distances are set to the same value.
                 if tuple or list, randomly picked in the interval `[yoffset[0], yoffset[1])`.
        p: probability that the image does this. Default value is 1.
    Return:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(xoffset, (list, tuple)):
        xoffset = np.random.randint(xoffset[0], xoffset[1])
    if yoffset is None:
        yoffset = xoffset
    elif isinstance(yoffset, (list, tuple)):
        yoffset = np.random.randint(yoffset[0], yoffset[1])
    return ImageChops.offset(image, xoffset, yoffset)


def pad(image, pad_value, pad_color=None, p=1):
    """Add border to the image
    
    Args:
        image: a PIL instance.
        pad_value: int or list or tuple, if int, pad same value with border, 
                   if list or tuple, len(pad_value)==2, left, top = right, bottom = pad_value
                   if list or tuple, len(pad_value)==4, left, top, right, bottom = pad_value
        pad_color: str or tuple or list or la.image.RGBMode, fill RGB color value, 
                   if str, hexadecimal color;
                   if len(pad_color) == 2, left_color, top_color = right_color, bottom_color = pad_color
                   if len(pad_color) == 3, left_color = top_color = right_color = bottom_color = pad_color
                   if len(pad_color) == 4, left_color, top_color, right_color, bottom_color = pad_color
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance. 
    """
    if np.random.uniform()>p:
        return image
    if isinstance(pad_value, (tuple, list)):
        if len(pad_value) == 2:
            left, top = right, bottom = pad_value
        elif len(pad_value) == 4:
            left, top, right, bottom = pad_value
        else:
            raise ValueError('`pad_value` value error.')
    else:
        left = top = right = bottom = pad_value
    
    if pad_color is None:
        left_color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))
        top_color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))
        right_color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))
        bottom_color = (np.random.randint(0, 255), np.random.randint(0, 255), np.random.randint(0, 255))
    elif isinstance(pad_color, dict):
        left_color = top_color = right_color = bottom_color = pad_color['rgb']
    elif isinstance(pad_color, (tuple, list)):
        pad_color = tuple([i['rgb'] if isinstance(i, dict) else i for i in pad_color])
        if len(pad_color) == 2:
            left_color, top_color = right_color, bottom_color = pad_color
        elif len(pad_color) == 3:
            left_color = top_color = right_color = bottom_color = pad_color
        elif len(pad_color) == 4:
            left_color, top_color, right_color, bottom_color = pad_color
        else:
            raise ValueError('`pad_value` value error.')
    else:
        left_color = top_color = right_color = bottom_color = pad_color

    width = left + image.size[0] + right
    height = top + image.size[1] + bottom

    out = Image.new(image.mode, (width, height), left_color)
    out.paste(image, (left, top))
    out.paste(Image.new(image.mode, (image.size[0] + right, top), top_color), (left, 0))
    out.paste(Image.new(image.mode, (right, top + image.size[1] + bottom), right_color), (left + image.size[0], top))
    out.paste(Image.new(image.mode, (left + image.size[0], bottom), bottom_color), (0, top + image.size[1]))
    return out


def channel_shuffle(image, p=1):
    """Random shuffle image channel.
    
    Args:
        image: a PIL instance.
        p: probability that the image does this. Default value is 1.
    returns: 
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    t = image.split()
    return Image.merge(image.mode, [t[i] for i in np.random.choice(list(range(len(t))), len(t), replace=False)])
        

def perspective(image, distortion_scale, fill_color=None, p=1):
    """Performs a random perspective transformation of the given image with a given probability. 

    Args:
        image: a PIL instance.
        distortion_scale: float, argument to control the degree of distortion and ranges from 0 to 1.
        fill_color: int or str or tuple or la.image.RGBMode, rgb color.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    (width, height) = image.size
    half_height = height // 2
    half_width = width // 2
    topleft = [np.random.randint(0, distortion_scale*half_width+1), 
               np.random.randint(0, distortion_scale*half_height+1)]
    topright = [np.random.randint(width - distortion_scale * half_width- 1, width),
                np.random.randint(0, distortion_scale * half_height + 1)]
    botright = [np.random.randint(width - distortion_scale * half_width - 1, width),
                np.random.randint(height - distortion_scale * half_height - 1, height)]
    botleft = [np.random.randint(0, distortion_scale * half_width + 1),
               np.random.randint(height - distortion_scale * half_height - 1, height)]
    
    startpoints = [[0, 0], [width - 1, 0], [width - 1, height - 1], [0, height - 1]]
    endpoints = [topleft, topright, botright, botleft]
    
    a_matrix = np.zeros([2 * len(startpoints), 8])
    for i, (p1, p2) in enumerate(zip(endpoints, startpoints)):
        a_matrix[2 * i, :] = np.array([p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]])
        a_matrix[2 * i + 1, :] = np.array([0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]])

    b_matrix = np.array(startpoints, dtype=np.float32).flatten()
    coeffs = np.linalg.lstsq(a_matrix, b_matrix, rcond=None)[0].tolist()
    fill_color = _fill_color(image, fill_color)
    return image.transform(image.size, Image.Transform.PERSPECTIVE, coeffs, 
                           Image.Resampling.NEAREST, fillcolor=fill_color)


def affine(image, angle=(-180, 180), center=(0.5,0.5), translate=(0, 0), scale=1., shear=(0,0), fill_color=None, p=1):
    """Apply affine transformation on the image.
    
    Args:
        image: a PIL instance.
        angle: int or float, rotation angle in degrees between -180 and 180. Set to 0 to deactivate rotations.
               if list or tuple, randomly picked in the interval `[angle[0], angle[1])`.
        center: center of rotation, xaxis and yaxis in [0,1], default is the center of the image.
                if int or float, xaxis=yaxis,
                if 2-tuple, (xaxis, yaxis), 
                if 4-tuple, xaxis in (center[0], center[1]) and yaxis in (center[2], center[3]).
        translate: post-rotate translation, xoffset and yoffset in [-1,1], see la.image.translate method.
                   if int or float, xoffset=yoffset,
                   if 2-tuple, (xoffset, yoffset), 
                   if 4-tuple, xoffset in (translate[0], translate[1]) and  yoffset in (translate[2], translate[3]).
        scale: float, scaling factor interval, should be positive.
               if list or tuple, randomly picked in the interval `[scale[0], scale[1])`.
        shear: Range of degrees to select from, xoffset and yoffset in [-360,360]. 
               if int or float, xoffset=yoffset,
               if 2-tuple, (xoffset, yoffset), 
               if 4-tuple, xoffset in (shear[0], shear[1]) and  yoffset in (shear[2], shear[3]).
        fill_color: int or str or tuple or la.image.RGBMode, rgb color. 
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(angle, (list, tuple)):
        angle = np.random.uniform(angle[0], angle[1])
        
    if isinstance(center, (int, float)):
        cx, cy = (int(center*image.size[0]), int(center*image.size[1]))
    elif isinstance(center, (list, tuple)):
        if len(center)==2:
            cx, cy = (int(center[0]*image.size[0]), int(center[1]*image.size[1]))
        elif len(center)==4:
            cx = int(np.random.uniform(center[0], center[1])*image.size[0])
            cy = int(np.random.uniform(center[2], center[3])*image.size[1])
        else:
            raise ValueError('`center` value format error.')
    else:
        raise ValueError('`center` value format error.')
    
    if isinstance(translate, (int, float)):
        tx, ty = int(translate*image.size[0]), int(translate*image.size[1])
    elif isinstance(translate, (list, tuple)):
        if len(translate)==2:
            tx, ty = int(translate[0]*image.size[0]), int(translate[1]*image.size[1])
        elif len(translate)==4:
            tx = int(np.random.uniform(translate[0], translate[1])*image.size[0])
            ty = int(np.random.uniform(translate[2], translate[3])*image.size[1])
        else:
            raise ValueError('`translate` value format error.')
    else:
        raise ValueError('`translate` value format error.')
        
    if isinstance(scale, (list, tuple)):
        scale = np.random.uniform(scale[0], scale[1])
    if scale <= 0.0:
        raise ValueError("Argument scale should be positive")

    if isinstance(shear, (int, float)):
        shear = (shear, shear)
    elif isinstance(shear, (list, tuple)):
        if len(shear)==4:
            shear = (np.random.uniform(shear[0], shear[1]), np.random.uniform(shear[2], shear[3]))
        elif len(shear)!=2:
            raise ValueError('`translate` value format error.')
    else:
        raise ValueError('`translate` value format error.')
    
    rot = math.radians(angle)
    sx = math.radians(shear[0])
    sy = math.radians(shear[1])

    a = math.cos(rot - sy) / math.cos(sy)
    b = -math.cos(rot - sy) * math.tan(sx) / math.cos(sy) - math.sin(rot)
    c = math.sin(rot - sy) / math.cos(sy)
    d = -math.sin(rot - sy) * math.tan(sx) / math.cos(sy) + math.cos(rot)

    matrix = [d, -b, 0.0, -c, a, 0.0]
    matrix = [x / scale for x in matrix]
    matrix[2] += matrix[0] * (-cx - tx) + matrix[1] * (-cy - ty)
    matrix[5] += matrix[3] * (-cx - tx) + matrix[4] * (-cy - ty)
    matrix[2] += cx
    matrix[5] += cy
    fill_color = _fill_color(image, fill_color)
    return image.transform(image.size, Image.Transform.AFFINE, matrix, 
                           Image.Resampling.NEAREST, fillcolor=fill_color)


def shear(image, xoffset=(-90, 90), yoffset=None, fill_color=None, p=1):
    """Apply affine shear on the image.
    
    Args:
        image: a PIL instance.
        xoffset: int or list ot tuple, The horizontal degrees, xoffset in [-360,360].
                 if tuple or list, randomly picked in the interval `[xoffset[0], xoffset[1])`
        yoffset: int or list ot tuple, The vertical degrees, yoffset in [-360,360]. 
                 If omitted, both distances are set to the same value.
                 if tuple or list, randomly picked in the interval `[yoffset[0], yoffset[1])`
        fill_color: int or str or tuple or la.image.RGBMode, rgb color. 
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(xoffset, (list, tuple)):
        xoffset = np.random.uniform(xoffset[0], xoffset[1])
    if yoffset is None:
        yoffset = xoffset
    elif isinstance(yoffset, (list, tuple)):
        yoffset = np.random.uniform(yoffset[0], yoffset[1])
    fill_color = _fill_color(image, fill_color)
    return affine(image, angle=(0,0), center=(0.5,0.5), translate=(0, 0), scale=1., 
                  shear=(xoffset, yoffset), fill_color=fill_color, p=1)


def rescale(image, xscale=(0.5,1.5), yscale=(0.5,1.5), fill_color=None, p=1):
    """Apply scaling on the x or y axis to input data.
    
    Args:
        image: a PIL instance.
        xscale: if int or float, width expansion and contraction, xscale should be positive.
                if tuple or list, randomly picked in the interval `[xscale[0], xscale[1])`.
        yscale: if int or float, height expansion and contraction, yscale should be positive.
                if tuple or list, randomly picked in the interval `[yscale[0], yscale[1])`.
        fill_color: int or str or tuple or la.image.RGBMode, rgb color.
        p: probability that the image does this. Default value is 1.
    Return:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(xscale, (list, tuple)):
        xscale = np.random.uniform(xscale[0], xscale[1])
    if isinstance(yscale, (list, tuple)):
        yscale = np.random.uniform(yscale[0], yscale[1])
    fill_color = _fill_color(image, fill_color)
    size = [int(image.size[0]*xscale), int(image.size[1]*yscale)]
    size1 = [min(image.size[0], size[0]), min(image.size[1], size[1])]
    box1 = [int(size[0]/2-size1[0]/2), int(size[1]/2-size1[1]/2)]
    box1 = box1 + [box1[0]+size1[0], box1[1]+size1[1]]
    box2 = [int(image.size[0]/2-size1[0]/2), int(image.size[1]/2-size1[1]/2)]
    box2 = box2 + [box2[0]+size1[0], box2[1]+size1[1]]
    image2 = Image.new(image.mode, image.size, fill_color)
    image2.paste(image.resize(size).crop(box1), box2)
    return image2


def jigsaw(image, size=(10,10), prob=0.1, p=1):
    """Move cells within images similar to jigsaw patterns.
    
    Args:
        image: a PIL instance.
        size: if int or float, xsize=ysize, numbers of jigsaw.
              if 2-tuple, (xsize, ysize), 
              if 4-tuple, xsize in (size[0], size[1]) and  ysize in (size[2], size[3]).
        prob: probability of every jigsaw being changed.
        p: probability that the image does this. Default value is 1.
    Return:
        a PIL instance.
    """
    if isinstance(size, (int, float)):
        size = (size, size)
    elif isinstance(size, (list, tuple)):
        if len(size)==4:
            size = (np.random.uniform(size[0], size[1]), np.random.uniform(size[2], size[3]))
        elif len(size)!=2:
            raise ValueError('`size` value format error.')
    else:
        raise ValueError('`size` value format error.')
    size = (max(2, int(size[0])), max(2, int(size[1])))
    
    w = image.size[0]//size[0]
    h = image.size[1]//size[1]
    axis = [[i*w, j*h, i*w+w, j*h+h] for i in range(size[0]) for j in range(size[1])]
    image2 = image.copy()
    s = np.random.choice(range(size[0]*size[1]), max(int(size[0]*size[1]*prob), 2), replace=False)
    s_index = np.random.choice(s, len(s), replace=False)
    for i,j in zip(s, s_index):
        image2.paste(image.crop(axis[i]), axis[j])
    return image2


from linora import gfile
from linora.utils._progbar import Progbar
from linora.data._utils import assert_dirs, get_file
from linora.data.datasets._config_path import Param

__all__ = ['download', 'list_datasets']


def list_datasets():
    """Support datasets list."""
    return list(Param.__dict__.keys())


def download(data_id, folder=None):
    """"Download datasets.
    
    Args:
        data_id: datasets id. see api: la.data.datasets.list_datasets()
        folder: str, Store the absolute path of the data directory.
    Returns:
        Store the absolute path of the data directory, example: `folder/mnist`.
    """
    task_path = assert_dirs(folder, data_id)
    try:
        data_list = Param.__getattribute__(data_id)
    except:
        raise ValueError(f'Not support datasets `{data_id}`.')
    
    if len(data_list)>1:
        bar = Progbar(len(data_list))
        for r, url in enumerate(data_list):
            root_file = gfile.path_join(task_path, url.split('/')[-1])
            get_file(url, root_file, verbose=0, retries=3, chunk_size=5120)
            bar.update(r+1)
    else:
        url = data_list[0]
        root_file = gfile.path_join(task_path, url.split('/')[-1])
        get_file(url, root_file, verbose=1, retries=3, chunk_size=5120)
    return task_path


from linora.metrics import distance
from linora.metrics._regression import *
from linora.metrics._classification import *
from linora.metrics._rank import *
from linora.metrics._metrics import *
from linora.metrics._image import *



import numpy as np

__all__ = ['Metrics']


class Metrics:
    """Generic metrics update class
    
    Args:
        function: function name, use la.mertics.*, function(y_true, y_pred)
    """
    def __init__(self, function):
        self.function = function
        self.result = []
    
    def compute(self, y_true, y_pred, sample_weight=None, **kwargs):
        """compute function value
        
        Args:
            y_true: pd.Series or array or list, ground truth (correct) labels.
            y_pred: pd.Series or array or list, predicted values.
            sample_weight: list or array of sample weight.
            **kwargs: function other params
        Return:
            function value
        """
        return self._function(y_true, y_pred, sample_weight, **kwargs)
        
    def update(self, y_true, y_pred, sample_weight=None, **kwargs):
        """update function value
        
        Args:
            y_true: pd.Series or array or list, ground truth (correct) labels.
            y_pred: pd.Series or array or list, predicted values.
            sample_weight: list or array of sample weight.
            **kwargs: function other params
        """
        self.result += [self._function(y_true, y_pred, sample_weight, **kwargs)]
        
    def accumulate(self, latest_num=-1, method='mean'):
        """compute function history value
        
        Args:
            latest_num: select the latest number of samples.
            method: samples compute method.['mean', 'sum', 'max', 'min']
        """
        if method=='mean':
            score = np.mean(self.result[-max(latest_num, 0):])
        elif method=='sum':
            score = np.sum(self.result[-max(latest_num, 0):])
        elif method=='max':
            score = np.max(self.result[-max(latest_num, 0):])
        elif method=='min':
            score = np.min(self.result[-max(latest_num, 0):])
        else:
            raise ValueError("`method` must one of `mean`, `sum`, `max`, `min`.")
        return score
    
    def reset():
        self.result = []
        
    def _function(self, y_true, y_pred, sample_weight, **kwargs):
        if sample_weight is None:
            return self.function(y_true, y_pred, **kwargs)
        else:
            return self.function(y_true, y_pred, sample_weight=sample_weight, **kwargs)



from linora.utils._config import Config

__all__ = ['Callback', 'CallbackList']


class Callback():
    """Abstract base class used to build new callbacks."""
    def __init__(self):
        self._params = Config()
        self.state = False
        self.checkpoint = False
        self.lr = None
        self._params.name = 'Callback'
        
    def update(self, batch, log=None):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        pass
    
    def _update(self, batch, log):
        self.update(batch, log)
        
        
class CallbackList():
    """Container abstracting a list of callbacks.
    
    Args:
        callback: List of Callback instances.
    """
    def __init__(self, callbacks=None):
        self._params = Config()
        self._params.callbacks = []
        if callbacks is not None:
            self._params.callbacks += callbacks if isinstance(callbacks, list) else [callbacks]
        self.state = False
        self.checkpoint = False
        self.lr = None
        for callback in self._params.callbacks:
            if 'LR' in callback._params.name:
                self.lr = callback._params.lr
            elif callback._params.name=='Callback':
                self.state = self.state or callback.state
                self.checkpoint = self.checkpoint or callback.checkpoint
                if callback.lr is not None:
                    self.lr = callback.lr
        self._params.name_list_state = ['EarlyStopping', 'TerminateOnNaN', 'TerminateOverfitted']
        self._params.name_list_checkpoint = ['ModelCheckpoint']
        
    def append(self, callback):
        """append callback.
        
        Args:
            callback: Callback instances.
        """
        self._params.callbacks.append(callback)
        if 'LR' in callback._params.name:
            self.lr = callback._params.lr
        
    def update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        for callback in self._params.callbacks:
            callback._update(batch, log)
            if callback._params.name in self._params.name_list_state:
                self.state = self.state or callback._params.state
            elif callback._params.name in self._params.name_list_checkpoint:
                self.checkpoint = self.checkpoint or callback._params.checkpoint
            elif 'LR' in callback._params.name:
                self.lr = callback._params.lr
            elif callback._params.name=='Callback':
                self.state = self.state or callback.state
                self.checkpoint = self.checkpoint or callback.checkpoint
                if callback.lr is not None:
                    self.lr = callback.lr
                


__all__ = ['datetime_extraction', 'datetime_transform']


def datetime_extraction(feature, extract, mode=0, name=None, config=None):
    """Extracts date and time features from datetime variables.
    
    The transformer supports the extraction of the following features:
    - 'year'
    - 'month'
    - 'day'
    - 'hour'
    - 'minute'
    - 'second'
    - 'date'
    - 'time'
    - 'weekday'
    - 'yearday'
    - 'quarter'
    - 'week'
    - 'days_in_month'
    - 'month_start'
    - 'month_end'
    - 'quarter_start'
    - 'quarter_end'
    - 'year_start'
    - 'year_end'
    - 'leap_year'
    
    Args:
        feature: pd.Series with datetime format, sample feature.
        extract: The list of date features to extract. 
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.        
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        if name is None:
            name = feature.name
        if isinstance(extract, str):
            extract = [extract]
        config = {'param':{'extract':extract, 'name':[f'{name}_{i}' for i in extract]},
                  'type':'datetime_extraction', 'variable':feature.name}
    if mode==2:
        return config
    else:
        var = ['year', 'month', 'day', 'hour', 'minute', 'second', 'date', 'time',
               'weekday', 'yearday', 'quarter', 'week', 'days_in_month',
               'month_start', 'month_end', 'quarter_start', 'quarter_end', 'year_start', 'year_end',
               'leap_year' ]
        if isinstance(extract, str):
            extract = [extract]
        for i in extract:
            assert i in var, f'extract variable `{i}` error, must be one of {var}.'
        data = []
        for i in extract:
            if i=='year':
                data.append(feature.dt.year)
            elif i=='month':
                data.append(feature.dt.month)
            elif i=='day':
                data.append(feature.dt.day)
            elif i=='hour':
                data.append(feature.dt.hour)
            elif i=='minute':
                data.append(feature.dt.minute)
            elif i=='second':
                data.append(feature.dt.second)
            elif i=='date':
                data.append(feature.dt.date)
            elif i=='time':
                data.append(feature.dt.time)
            elif i=='weekday':
                data.append(feature.dt.day_of_week)
            elif i=='yearday':
                data.append(feature.dt.day_of_year)
            elif i=='quarter':
                data.append(feature.dt.quarter)
            elif i=='week':
                data.append(feature.dt.isocalendar().week)
            elif i=='days_in_month':
                data.append(feature.dt.days_in_month)
            elif i=='month_start':
                data.append(feature.dt.is_month_start)
            elif i=='month_end':
                data.append(feature.dt.is_month_end)
            elif i=='quarter_start':
                data.append(feature.dt.is_quarter_start)
            elif i=='quarter_end':
                data.append(feature.dt.is_quarter_end)
            elif i=='year_start':
                data.append(feature.dt.is_year_start)
            elif i=='year_end':
                data.append(feature.dt.is_year_end)
            elif i=='leap_year':
                data.append(feature.dt.is_leap_year)
            
        data = pd.concat(data, axis=1)
        data.columns = config['param']['name']
        return data if mode else (data, config)


def datetime_transform(feature, format_in, format_out='timestamp', mode=0, name=None, config=None):
    """Time format transform.
    
    The transformer supports the time format of the following features:
    
    - 'timenumber:s', 's' id unit, one of ['s', 'ms', 'us', 'ns']:
        1679455111, 1679455111.0958722
    - 'strtime:format', format is "%Y-%m-%d %H:%M:%S":
        %y Two-digit Year Representation (00-99)
        %Y Four-digit Year Representation (000-9999)
        %m month (01-12)
        %d One day in months (0-31)
        %h Hours in 24-hour clock (0-23)
        %I 12-hour Hours (01-12)
        %M minutes (00=59)
        %S seconds (00-59)
        %a local simplified week name
        %A local full week name
        %b Local Simplified Month Name
        %B Local full month name
        %c local corresponding date representation and time representation
        %j One day in year (001-366)
        %p equivalent of local a.m. or p.m.
        %U Number of weeks in a year (00-53) Sunday is the beginning of the week.
        %w week (0-6), Sunday is the beginning of the week.
        %W Number of weeks in a year (00-53) Monday is the beginning of the week.
        %x local corresponding date representation
        %X local corresponding time representation
        %Z Name of the current time zone
        eg:
        - "%Y-%m-%d %H:%M:%S": '2023-01-01 02:03:56'
        - "%A %Y-%m %H:%M:%S": 'Thursday 2023-01 02:03:56'
    - 'timestamp':
        Timestamp(2023-01-01 02:03:56)
    - 'datetime'
        datetime.datetime(2023, 3, 23, 8, 9, 59, 733760)
    
    Args:
        feature: pd.Series with datetime format, sample feature.
        format_in: input time format.
        format_out: output time format.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.        
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'format_in':format_in, 'format_out':format_out,
                          'name':feature.name if name is None else name},
                  'type':'datetime_transform', 'variable':feature.name}
    if mode==2:
        return config
    else:
        format_in_mode = format_in.split(':')[0]
        format_in_param = format_in[len(format_in_mode)+1:]
        format_out_mode = format_out.split(':')[0]
        format_out_param = format_out[len(format_out_mode)+1:]

        if format_in_mode=='timenumber':
            if format_in_param=='':
                format_in_param = 's'
            t = pd.to_datetime(feature, unit=format_in_param)
        elif format_in_mode=='strtime':
            if format_in_param=='':
                format_in_param = None
            t = pd.to_datetime(feature, format=format_in_param)
        elif format_in_mode=='datetime':
            t = pd.to_datetime(feature)
        elif format_in_mode=='timestamp':
            t = feature
        else:
            raise ValueError('`format_in` value error.')

        if format_out_mode=='timenumber':
            t = t.map(lambda x: x.timestamp())
        elif format_out_mode=='strtime':
            if format_out_param=='':
                format_out_param = "%Y-%m-%d %H:%M:%S"
            t = t.dt.strftime(format_out_param)
        elif format_out_mode=='datetime':
            t = t.map(lambda x: x.to_pydatetime())
        elif format_out_mode=='timestamp':
            pass
        else:
            raise ValueError('`format_out` value error.')
        return t.rename(config['param']['name']) if mode else (t.rename(config['param']['name']), config)
 



from linora.utils._config import Config

Param = Config()

Param.mnist = [
    'https://apache-mxnet.s3-accelerate.dualstack.amazonaws.com/gluon/dataset/mnist/train-labels-idx1-ubyte.gz',
    'https://apache-mxnet.s3-accelerate.dualstack.amazonaws.com/gluon/dataset/mnist/train-images-idx3-ubyte.gz',
    'https://apache-mxnet.s3-accelerate.dualstack.amazonaws.com/gluon/dataset/mnist/t10k-labels-idx1-ubyte.gz',
    'https://apache-mxnet.s3-accelerate.dualstack.amazonaws.com/gluon/dataset/mnist/t10k-images-idx3-ubyte.gz',
]

Param.mnist_fashion = [
    'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz',
    'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz',
    'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz',
    'http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz',
]

Param.mnist_kannada = [
    'https://github.com/Hourout/datasets/releases/download/0.0.1/kannada_MNIST.zip',
]

Param.mnist_kuzushiji10 = [
    'http://codh.rois.ac.jp/kmnist/dataset/kmnist/kmnist-train-imgs.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/kmnist/kmnist-train-labels.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/kmnist/kmnist-test-imgs.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/kmnist/kmnist-test-labels.npz',
]

Param.mnist_kuzushiji49 = [
    'http://codh.rois.ac.jp/kmnist/dataset/k49/k49-train-imgs.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/k49/k49-train-labels.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/k49/k49-test-imgs.npz',
    'http://codh.rois.ac.jp/kmnist/dataset/k49/k49-test-labels.npz',
]

Param.mnist_kuzushiji_kanji = [
    "http://codh.rois.ac.jp/kmnist/dataset/kkanji/kkanji.tar",
]

Param.mnist_tibetan = [
    'https://raw.githubusercontent.com/Hourout/datasets/master/TibetanMNIST/TibetanMNIST_28_28_01.csv',
    'https://raw.githubusercontent.com/Hourout/datasets/master/TibetanMNIST/TibetanMNIST_28_28_02.csv',
]

Param.lfw = [
    "http://vis-www.cs.umass.edu/lfw/lfw.tgz"
]








import time
import warnings

import psutil
import pandas as pd
from IPython.display import display_html, clear_output


__all__ = ['process_monitor']


def memory_transform(data):
    for r,i in enumerate(['KB', 'MB', 'GB', 'TB'], start=1):
        t = data/1024**r
        if t<1024:
            break
    return f"{round(t,1)}{i}"

def process_monitor(display=True, monitor_memory=True, monitor_gpu=True, monitor_disk=False, monitor_pid=False, 
                    sleep=5, sleep_total=None, process_min_bytes=1024**2):
    pynvml_state = False
    try:
        if monitor_gpu:
            import pynvml
            pynvml.nvmlInit()
            pynvml_state = True
    except:
        warnings.warn("should install 'pynvml' package.")
    
    sleep_count = 0
    while 1:
        try:
            process = {}
            if monitor_memory:
                virtual_memory = psutil.virtual_memory()
                process['memory_total'] = memory_transform(virtual_memory.total)
                process['memory_used'] = memory_transform(virtual_memory.used)
                process['memory_percent'] = f"{virtual_memory.percent}%"
                process['cpu_count_physics'] = psutil.cpu_count()
                process['cpu_count_logical'] = psutil.cpu_count(logical=False)

            if monitor_disk:
                process['disk'] = []
                for i in psutil.disk_partitions():
                    t = psutil.disk_usage(i.mountpoint)
                    process['disk'].append({'device':i.device, 'mountpoint':i.mountpoint, 
                                            'total':memory_transform(t.total), 'used':memory_transform(t.used), 
                                            'percent':f"{t.percent}%", 'fstype':i.fstype})

            if monitor_pid:
                process['pid'] = []
                for proc in psutil.process_iter():
                    try:
                        if proc.cmdline():
                            memory_used = proc.memory_full_info().uss
                            if memory_used>process_min_bytes:
                                data = proc.as_dict()
                                data = {i:data[i] for i in ['pid', 'ppid', 'username', 'name', 'create_time', 'memory_percent', 'cmdline']}
                                data['create_time'] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(data['create_time']))
                                data['memory_used'] = memory_transform(memory_used)
                                process['pid'].append(data)
                    except:
                        pass

            if monitor_gpu and pynvml_state:
                process['gpu_memory'] = []
                process['gpu_pid'] = []
                for i in range(pynvml.nvmlDeviceGetCount()):
                    data = {'Kernel':f'GPU:{i}'}
                    handle = pynvml.nvmlDeviceGetHandleByIndex(i)
                    data['Name'] = pynvml.nvmlDeviceGetName(handle).decode()
                    meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
                    data['Memory-Usage'] = f"{memory_transform(meminfo.used)} / {memory_transform(meminfo.total)}"
                    data['Temperature'] = f'{pynvml.nvmlDeviceGetTemperature(handle,0)}℃'
                    data['FanSpeed'] = f'{pynvml.nvmlDeviceGetFanSpeed(handle)}%'
                    process['gpu_memory'].append(data)

                    for j in pynvml.nvmlDeviceGetComputeRunningProcesses(handle):
                        data = {'Kernel':f'GPU:{i}'}
                        data['PID'] = j.pid
                        data['Memory-Usage'] = f"{memory_transform(j.usedGpuMemory)}"
                        data['Process-name'] = pynvml.nvmlSystemGetProcessName(j.pid).decode()
                        process['gpu_pid'].append(data)

            if not display:
                return process
            clear_output(wait=True)
            if monitor_memory:
                temp = ['memory_total', 'memory_used', 'memory_percent', 'cpu_count_physics', 'cpu_count_logical']
                display_html(pd.DataFrame({i:[process[i]] for i in temp}))
            if monitor_gpu and pynvml_state:
                display_html(pd.DataFrame(process['gpu_memory']))
                display_html(pd.DataFrame(process['gpu_pid']))
            if monitor_disk:
                display_html(pd.DataFrame(process['disk']))
            if monitor_pid:
                display_html(pd.DataFrame(process['pid']))
        except:
            pass
        time.sleep(sleep)
        sleep_count += sleep
        if sleep_total is not None:
            if sleep_total>sleep_count:
                break




import itertools
from functools import reduce
from collections import defaultdict

import pandas as pd
import numpy as np

__all__ = ['categorical_count', 'categorical_crossed','categorical_encoder', 
           'categorical_hash', 'categorical_hist',
           'categorical_onehot_binarizer', 'categorical_onehot_multiple',
           'categorical_rare', 'categorical_regress', 'categorical_woe'
          ]


def categorical_count(feature, mode=0, normalize=True, abnormal_value=0, miss_value=0, name=None, config=None):
    """Count or frequency of conversion category variables.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        normalize: bool, If True then the object returned will contain the relative frequencies of the unique values.
        abnormal_value: int or float, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':feature.value_counts(normalize).to_dict(),
                           'normalize':normalize, 'abnormal_value':abnormal_value, 
                           'miss_value':miss_value, 'name':feature.name if name is None else name},
                  'type':'categorical_count', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = {**config['param']['feature_scale'], **{i:config['param']['abnormal_value'] for i in feature.unique().tolist() if i not in config['param']['feature_scale']}}
        t = feature.replace(scale).fillna(config['param']['miss_value']).rename(config['param']['name'])
        return t if mode else (t, config)


def categorical_crossed(feature_list, mode=0, hash_bucket_size=3, name=None, config=None):
    """Crossed categories and hash labels with value between 0 and hash_bucket_size-1.
    
    Args:
        feature_list: pd.Series list, sample feature list.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        hash_bucket_size: int, number of categories that need hash.
        name: str, output feature name.
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature_list` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'hash_bucket_size':hash_bucket_size, 
                           'name':'_'.join(name)+'_crossed' if name is None else name}, 
                  'type':'categorical_crossed', 'variable':[i.name for i in feature_list]}
    if mode==2:
        return config
    else:
        t = reduce(lambda x,y:x+y, [i.fillna('').astype(str) for i in feature_list]).map(lambda x:hash(x)).rename(config['param']['name'])%config['param']['hash_bucket_size']
        return t if mode else (t, config)


def categorical_encoder(feature, mode=0, abnormal_value=-1, miss_value=-1, name=None, config=None):
    """Encode labels with value between 0 and n_classes-1.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':{j:i for i,j in feature.drop_duplicates().reset_index(drop=True).to_dict().items()},
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_encoder', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = {**config['param']['feature_scale'], **{i:config['param']['abnormal_value'] for i in feature.unique().tolist() if i not in config['param']['feature_scale']}}
        t = feature.replace(scale).fillna(config['param']['miss_value']).rename(config['param']['name'])
        return t if mode else (t, config)


def categorical_hash(feature, mode=0, hash_bucket_size=3, name=None, config=None):
    """Hash labels with value between 0 and hash_bucket_size-1.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        hash_bucket_size: int, number of categories that need hash.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'hash_bucket_size':hash_bucket_size, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_hash', 'variable':feature.name}
    if mode==2:
        return config
    else:
        t = feature.fillna('').astype(str).map(lambda x:hash(x)).rename(config['param']['name'])%config['param']['hash_bucket_size']
        return t if mode else (t, config)


def categorical_hist(feature, label, mode=0, abnormal_value=0, miss_value=0, name=None, config=None):
    """Hist labels with value counts prob.
           
    Args:
        feature: pd.Series, sample feature.
        label: pd.Series, sample categorical label.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid, but `label` must be passed in.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':None,
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_hist', 'variable':feature.name, 'label':label.name}
        t = pd.concat([feature, label], axis=1).groupby([feature.name])[label.name].value_counts(normalize=True).unstack()
        t.columns = [config['param']['name']+'_'+str(i) for i in t.columns]
        config['param']['feature_scale'] = t.reset_index().to_dict()
        
    if mode==2:
        return config
    else:
        t = (feature.to_frame()
             .merge(pd.DataFrame(config['param']['feature_scale'])
                    .fillna(config['param']['miss_value']), on=feature.name, how='left')
             .drop([feature.name], axis=1).fillna(config['param']['abnormal_value']).rename(config['param']['name']))
        return t if mode else (t, config)


def categorical_onehot_binarizer(feature, mode=0, abnormal_value=0, miss_value=0, name=None, config=None):
    """Transform between iterable of iterables and a multilabel format, sample is simple categories.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':feature.dropna().drop_duplicates().tolist(),
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_onehot_binarizer', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = feature.dropna().drop_duplicates().tolist()
        scale_dict = {i:'temp_str' for i in set.difference(set(scale), set(config['param']['feature_scale']))}
        if feature.dtype.char=='O':
            t = pd.get_dummies(feature.replace(scale_dict), prefix=config['param']['name'], dtype='int8', dummy_na=True)
        else:
            t = pd.get_dummies(feature.replace(scale_dict).astype(str), prefix=config['param']['name'], dtype='int8', dummy_na=True)
        
        for i in set.difference(set(config['param']['feature_scale']), set(scale)):
            if config['param']['name']+'_'+str(i) not in t.columns:
                t[config['param']['name']+'_'+str(i)] = 0
                
        if f"{config['param']['name']}_temp_str" in t.columns:
            t.loc[t[f"{config['param']['name']}_temp_str"]==1, :] = config['param']['abnormal_value']
            t = t.drop([f"{config['param']['name']}_temp_str"], axis=1)
            
        if f"{config['param']['name']}_nan" in t.columns:
            t.loc[t[f"{config['param']['name']}_nan"]==1, :] = config['param']['miss_value']
            t = t.drop([f"{config['param']['name']}_nan"], axis=1)
        t = t[[config['param']['name']+'_'+str(i) for i in config['param']['feature_scale']]]
        return t if mode else (t, config)


def categorical_onehot_multiple(feature, mode=0, abnormal_value=0, miss_value=0, name=None, config=None):
    """Transform between iterable of iterables and a multilabel format, sample is multiple categories.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'feature_scale':list(set(itertools.chain.from_iterable(feature.dropna()))),
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_onehot_multiple', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = list(set(itertools.chain.from_iterable(feature.fillna('_'))))        
        class_mapping = defaultdict(int)
        class_mapping.default_factory = class_mapping.__len__
        [class_mapping[i] for i in scale]
        col = [class_mapping[i] for i in itertools.chain.from_iterable(feature.fillna('_'))]
        row = [i for i in itertools.chain.from_iterable(map(lambda x,y:[x]*len(y), range(len(feature)), feature.fillna('_')))]
        t = np.zeros([max(row)+1, len(class_mapping)], dtype='int8')
        t[row, col] = 1
        t = pd.DataFrame(t, columns=[config['param']['name']+'_'+str(i) for i in scale], index=feature.index)
    
        for i in set.difference(set(config['param']['feature_scale']), set(scale)):
            if config['param']['name']+'_'+str(i) not in t.columns:
                t[config['param']['name']+'_'+str(i)] = 0

        if f"{config['param']['name']}__" in t.columns:
            t.loc[t[f"{config['param']['name']}__"]==1, :] = config['param']['miss_value']
            t = t.drop([f"{config['param']['name']}__"], axis=1)
        
        t = t[[config['param']['name']+'_'+str(i) for i in config['param']['feature_scale']]]
        return t if mode else (t, config)


def categorical_rare(feature, mode=0, p=0.05, min_num=None, max_num=None, abnormal_value=-1, miss_value=-1, name=None, config=None):
    """Groups rare or infrequent categories in a new category called “Rare”, or any other name entered by the user.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        p: The minimum frequency a label should have to be considered frequent. Categories with frequencies lower than tol will be grouped.
        min_num: The minimum number of categories a variable should have for the encoder to find frequent labels. 
            If the variable contains less categories, all of them will be considered frequent.
        max_num: The maximum number of categories that should be considered frequent. 
            If None, all categories with frequency above the tolerance (tol) will be considered frequent. 
            If you enter 5, only the 4 most frequent categories will be retained and the rest grouped.
        abnormal_value: int or float, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        t = feature.value_counts(True)
        if min_num is None:
            min_num = len(t[t>p])
        if max_num is None:
            max_num = max(min_num, len(t[t>p])+1)
        k = 1 if len(t[t<=p])>0 else 0
        if len(t)<min_num:
            feature_scale = {i:r for r, i in enumerate(t.index)}
        elif len(t[t>p])+k<min_num:
            feature_scale = {**{i:r for r, i in enumerate(t.iloc[:min_num-1].index)},
                             **{i:min_num-1 for i in t.iloc[min_num-1:].index}}
        elif len(t[t>p])+k>max_num:
            feature_scale = {**{i:r for r, i in enumerate(t.iloc[:max_num-1].index)},
                             **{i:max_num-1 for i in t.iloc[max_num-1:].index}}
        else:
            feature_scale = {**{i:r for r, i in enumerate(t[t>p].index)},
                             **{i:len(t[t>p]) for i in t[t<=p].index}}
            
        config = {'param':{'feature_scale':feature_scale,
                           'p':p, 'min_num':min_num, 'max_num':max_num,
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_rare', 'variable':feature.name}
    if mode==2:
        return config
    else:
        scale = {**config['param']['feature_scale'], **{i:config['param']['abnormal_value'] for i in feature.unique().tolist() if i not in config['param']['feature_scale']}}
        t = feature.replace(scale).fillna(config['param']['miss_value']).rename(config['param']['name'])
        return t if mode else (t, config)


def categorical_regress(feature, label, mode=0, method='mean', abnormal_value='mean', miss_value='mean', name=None, config=None):
    """Regress labels with value counts prob.
    
    Args:
        feature: pd.Series, sample feature.
        label: pd.Series, sample regress label.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        method: 'mean' or 'median'
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid, but `label` must be passed in.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'method':method,
                           'feature_scale':pd.concat([feature, label], axis=1).groupby([feature.name])[label.name].agg(method).to_dict(),
                           'abnormal_value':label.mean() if abnormal_value=='mean' else label.median(), 
                           'miss_value':label.mean() if miss_value=='mean' else label.median(), 
                           'name':feature.name if name is None else name},
                  'type':'categorical_regress', 'variable':feature.name, 'label':label.name}
    if mode==2:
        return config
    else:
        scale = {**config['param']['feature_scale'], **{i:config['param']['abnormal_value'] for i in feature.unique().tolist() if i not in config['param']['feature_scale']}}
        t = feature.replace(scale).fillna(config['param']['miss_value']).rename(config['param']['name'])
        return t if mode else (t, config)

    
def categorical_woe(feature, label, mode=0, pos_label=1, abnormal_value=-1, miss_value=-1, name=None, config=None):
    """Calculate series woe value
    
    Args:
        feature: pd.Series, shape (n_samples,) x variable, model feature.
        label: pd.Series, sample categorical label.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        pos_label: int, default=1, positive label value.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
        miss_value: int or float, if feature values are missing, return `miss_value`.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid, but `label` must be passed in.
    Return:
        Refer to params `mode` explanation.
    """
    if config is None:
        t = pd.DataFrame({'label':label, 'feature':feature})
        assert t.label.nunique()==2, "`y_true` should be binary classification."
        label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
        t['label'] = t.label.replace(label_dict)
        corr = t.label.sum()/(t.label.count()-t.label.sum())
        t = t.groupby(['feature']).label.apply(lambda x:np.log(x.sum()/(x.count()-x.sum())/corr))
        
        config = {'param':{'pos_label':pos_label, 'feature_scale':t.to_dict(),
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':feature.name if name is None else name},
                  'type':'categorical_woe', 'variable':feature.name, 'label':label.name}
    if mode==2:
        return config
    else:
        scale = {**config['param']['feature_scale'], **{i:config['param']['abnormal_value'] for i in feature.unique().tolist() if i not in config['param']['feature_scale']}}
        t = feature.replace(scale).fillna(config['param']['miss_value']).rename(config['param']['name'])
        return t if mode else (t, config)
    




import matplotlib.pyplot as plt

from linora.utils._config import Config

__all__ = ['Options']

Options = Config()
Options.cmap = Config(**{'viridis':'viridis', 'jet':'jet', 'cubehelix':'cubehelix', 'rdbu':'RdBu', 'binary':'binary'})
Options.dash_capstyle = Config(**{'butt': 'butt', 'projecting': 'projecting', 'round': 'round'})
Options.dash_joinstyle = Config(**{'miter': 'miter', 'round': 'round', 'bevel': 'bevel'})
Options.linelink = Config(**{'steps':'steps', 'steps_pre':'steps-pre', 'steps_mid':'steps-mid', 'steps_post':'steps-post'})
Options.linestyle = Config(**{'solid':'-', 'dashed':'--', 'dashdot':'-.', 'dotted':':'})
Options.fontsize = Config(**{
    'small_xx':'xx_small', 
    'small_x':'x-small', 
    'small':'small', 
    'medium':'medium', 
    'large':'large', 
    'large_x':'x-large', 
    'large_xx':'xx-large'})
Options.fontweight = Config(**{
    'book': 'book', 
    'normal': 'normal',
    'bold': 'bold',
    'demi': 'demi',
    'semibold': 'semibold',
    'roman': 'roman',
    'black': 'black',
    'extra bold': 'extra bold',
    'light': 'light',
    'regular': 'regular',
    'demibold': 'demibold',
    'medium': 'medium',
    'ultralight': 'ultralight',
    'heavy': 'heavy'})
Options.mark = Config(**{
    'point': '.',
    'pixel': ',',
    'circle': 'o',
    'triangle_down': 'v',
    'triangle_up': '^',
    'triangle_left': '<',
    'triangle_right': '>',
    'tri_down': '1',
    'tri_up': '2',
    'tri_left': '3',
    'tri_right': '4',
    'octagon': '8',
    'square': 's',
    'pentagon': 'p',
    'star': '*',
    'hexagon1': 'h',
    'hexagon2': 'H',
    'plus': '+',
    'x': 'x',
    'diamond': 'D',
    'thin_diamond': 'd',
    'vline': '|',
    'hline': '_',
    'plus_filled': 'P',
    'x_filled': 'X',
    'tickleft': 0,
    'tickright': 1,
    'tickup': 2,
    'tickdown': 3,
    'caretleft': 4,
    'caretright': 5,
    'caretup': 6,
    'caretdown': 7,
    'caretleftbase': 8,
    'caretrightbase': 9,
    'caretupbase': 10,
    'caretdownbase': 11})
Options.markfill = Config(**{'full': 'full', 'left': 'left', 'right': 'right', 'bottom': 'bottom', 'top': 'top'})
Options.solid_capstyle = Options.dash_capstyle
Options.solid_joinstyle = Options.dash_joinstyle
Options.xloc = Config(**{'left':'left', 'center':'center', 'right':'right'})


Options.theme = Config(**{i.replace('-', '_'):i for i in plt.style.available})
Options.axis = Config(**{'on':'on', 'off':'off', 'equal':'equal', 'scaled':'scaled', 
                         'tight':'tight', 'auto':'auto', 'image':'image', 'square':'square'})

Options.color = {'green':['#9eccab', '#8cc269', '#68b88e', '#5dbe8a', '#5bae23', '#41b349', '#20894d',
                          '#1a6840', '#69a794', '#2c9678', '#12aa9c', '#57c3c2'],
                 'pink':['#eeb8c3', '#f0a1a8', '#ec8aa4', '#ec7696', '#ef82a0', '#de7897', '#e77c8e',
                         '#ed9db2', '#f07c82', '#ea517f', '#eb507e', '#ce5777',
                         '#eeaa9c', '#f17666', '#ed4845', '#f25a47', '#f03f24', '#ed3333', '#f43e06', 
                        '#eb261a', '#e60012', '#4d2517', '#9e2a22', '#77171c', ],
                 'blue':['#b4d9ef', '#78bbdd', '#6f94cd', '#63bbd0', '#158bb8', '#2775b6', '#0d67bf', 
                         '#1661ab', '#11659a', '#4656b7', '#2e317c', '#144a74', ],
                 'purple':['#c8adc4', '#c08eaf', '#c06f98', '#ad6598', '#806d9e', '#983680', '#815c94', 
                           '#813c85', '#9b1e64', '#7e2065', '#8b2671', '#7e1671', ],
                 'yellow':['#f7de98', '#e1d384', '#f8d86a', '#fed71a', '#e2d849', '#fccb16', '#feba07', 
                           '#edc300', '#d3a237', '#ffa60f', '#c57917', '#ff9900', ],
                 'grey':['#cfccc9', '#d4c4b7', '#b2bbbe', '#bdaead', '#b89485', '#9a8878', '#856d72', 
                         '#617172', '#8b614d', '#624941', '#475164', '#363433', ]
                }




from linora.data import Dataset
from linora.data import datasets
from linora.data import TextLineDataset
from linora.data._compress import *
from linora.data._utils import *
from linora.data._sql_io_dataset import *


import numpy as np

def _sample_weight(y_true, sample_weight):
    if sample_weight is None:
        sample_weight = np.ones(len(y_true))
    elif isinstance(sample_weight, dict):
        sample_weight = np.array([sample_weight[i] for i in y_true])
    else:
        sample_weight = np.array(sample_weight)
    return sample_weight/sample_weight.sum()*len(sample_weight)


from matplotlib import patches


class Wedge():
    def add_wedge(self, name, center, r, theta1, theta2, width=None, **kwargs):
        """Wedge shaped patch.
        
        A wedge centered at *x*, *y* center with radius *r* that
        sweeps *theta1* to *theta2* (in degrees).  If *width* is given,
        then a partial wedge is drawn from inner radius *r* - *width*
        to outer radius *r*.
        
        Args:
            alpha: scalar or None
            animated: bool
            antialiased or aa: unknown
            capstyle: `.CapStyle` or {'butt', 'projecting', 'round'}
            clip_box: `.Bbox`
            clip_on: bool
            clip_path: Patch or (Path, Transform) or None
            color: color
            contains: unknown
            edgecolor or ec: color or None or 'auto'
            facecolor or fc: color or None
            fill: bool
            gid: str
            hatch: {'/', '\\', '|', '-', '+', 'x', 'o', 'O', '.', '*'}
            in_layout: bool
            joinstyle: `.JoinStyle` or {'miter', 'round', 'bevel'}
            label: object
            linestyle or ls: {'-', '--', '-.', ':', '', (offset, on-off-seq), ...}
            linewidth or lw: float or None
            picker: None or bool or float or callable
            rasterized: bool
            sketch_params: (scale: float, length: float, randomness: float)
            snap: bool or None
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]
        kwargs['center'] = center
        kwargs['r'] = r
        kwargs['theta1'] = theta1
        kwargs['theta2'] = theta2
        kwargs['width'] = width
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['transform'] = 'ax'
        self._params.ydata[name]['plotmode'] = 'wedge'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_wedge
        return self
    
    def _execute_plot_wedge(self, fig, ax, i, j):
        poly = patches.Wedge(**j['kwargs'])
        if j['transform']=='ax':
            ax.add_patch(poly)
        else:
            fig.add_artist(poly)


from linora.feature_column._numerical import *

class FeatureNumerical(object):
    def __init__(self):
        self.pipe = {}
        
    def numerical_binarizer(self, variable, method='mean', name=None, keep=True):
        """Encode labels with value between 0 and 1.

        Args:
            variable: str, feature variable name.
            method: default 'mean', one of 'mean' or 'median' or float or int.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'method':method, 'name':variable if name is None else name},
                  'type':'numerical_binarizer', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_bucketized(self, variable, boundaries, miss_pad=-1, score=None, miss_score=None, method=1, name=None, keep=True):
        """feature bucket.

        if method is True:
            Buckets include the right boundary, and exclude the left boundary. 
            Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.], (0., 1.], (1., 2.], and (2., +inf).
        else:
            Buckets include the left boundary, and exclude the right boundary. 
            Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.), [0., 1.), [1., 2.), and [2., +inf).

        Args:
            variable: str, feature variable name.
            boundaries: list, A sorted list or tuple of floats specifying the boundaries.
            miss_pad: int, default -1, feature fillna value.
            score: None, A score list or tuple of floats specifying the boundaries.
            miss_score: int or float, None, score fillna value.
            method: True.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'boundaries':boundaries, 'miss_pad':miss_pad,
                           'score':score, 'miss_score':miss_score, 'method':method,
                           'name':variable if name is None else name},
                  'type':'numerical_bucketized', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_combine(self, variable, function, name=None, keep=True):
        """feature combine transform.

        Args:
            variable: str of list, feature variable name.
            function: function or str, Function to use for aggregating the data. 
                If a function, must either work when passed a DataFrame or when passed to DataFrame.apply.
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'func':func, 
                          'name':'combine_'+'_'.join(variable) if name is None else name},
                  'type':'numerical_combine', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_cyclical(self, variable, name=None, keep=True):
        """feature cyclical transform.

        applies cyclical transformations to numerical variables, returning 2 new features per variable.
        according to:
            var_sin = sin(variable * (2. * pi / max_value))
            var_cos = cos(variable * (2. * pi / max_value))

        where max_value is the maximum value in the variable, and pi is 3.14…

        Args:
            variable: str, feature variable name.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'name':variable if name is None else name},
                  'type':'numerical_cyclical', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self

    def numerical_math(self, variable, method=['log'], log='e', c='auto', power=0.5, name=None, keep=True):
        """feature math transform.

        using:
        - log
        - power
        - arcsin
        - reciprocal
        - BoxCox
        - YeoJohnson

        log:
            x = log(x+c)
            if c='auto', c = x.min().abs() + 1
            if x<0, x=0

            if log='e', x = log-e(x+c)
            if log='2', x = log-2(x+c)
            if log='10', x = log-10(x+c)

        power:
            x = x^power
            if power<1, x[x<=0] = 1

        arcsin:
            x = arcsin(sqrt(x))
            x must be 0~1

        reciprocal:
            x = 1/x
            if x=0, x=0

        BoxCox:
            applies the BoxCox transformation to numerical variables.
            The Box-Cox transformation is defined as:
            - T(Y)=(Y exp(λ)−1)/λ if λ!=0
            - log(Y) otherwise
            where Y is the response variable and λ is the transformation parameter. 
            λ varies, typically from -5 to 5. 
            In the transformation, all values of λ are considered and the optimal value for a given variable is selected.
            `Box and Cox. “An Analysis of Transformations”. Read at a RESEARCH MEETING, 1964.`

        YeoJohnson:
            applies the Yeo-Johnson transformation to the numerical variables.
            `Yeo, In-Kwon and Johnson, Richard (2000). 
            A new family of power transformations to improve normality or symmetry. Biometrika, 87, 954-959.`

        Args:
            variable: str, feature variable name.
            method: list, must be one or more of ['log', 'power', 'arcsin', 'reciprocal', 'BoxCox', 'YeoJohnson']
            log: Indicates if the natural or base 10 logarithm should be applied. Can take values ‘e’ or ‘10’, '2'.
            c: The constant C to add to the variable before the logarithm, i.e., log(x + C).
            power: The power (or exponent).
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'method':method, 'log':log, 'c':c, 'power':power,
                           'name':variable if name is None else name},
                  'type':'numerical_math', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_padding(self, variable, method='mean', name=None, keep=True):
        """feature fillna method.

        Args:
            variable: str, feature variable name.
            method: default 'mean', one of 'mean' or 'median' or float or int.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'method':method, 'name':variable if name is None else name},
                  'type':'numerical_padding', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_outlier(self, variable, method='norm', delta=0.9545, tail='right', name=None, keep=True):
        """feature clip outlier.

        Caps maximum and/or minimum values of a variable at automatically determined values, and optionally adds indicators
        The extreme values beyond which an observation is considered an outlier are determined 

        using:
        - norm
        - a Gaussian approximation
        - the inter-quantile range proximity rule (IQR)
        - MAD-median rule (MAD)
        - percentiles

        norm limits:
            right tail: mean + ppf((1-delta)/2)* std
            left tail: mean - ppf(delta+(1-delta)/2)* std

        Gaussian limits(Delta is recommended to be 3):
            right tail: mean + 3* std
            left tail: mean - 3* std

        IQR limits(Delta is recommended to be 3):
            right tail: 75th quantile + 3* IQR
            left tail: 25th quantile - 3* IQR
        where IQR is the inter-quartile range: 75th quantile - 25th quantile.

        MAD limits(Delta is recommended to be 3):
            right tail: median + 3* MAD
            left tail: median - 3* MAD
        where MAD is the median absoulte deviation from the median.

        percentiles:
            right tail: 95th percentile
            left tail: 5th percentile

        You can select how far out to cap the maximum or minimum values with the parameter 'delta'.
        If capping_method='gaussian' delta gives the value to multiply the std.
        If capping_method='iqr' delta is the value to multiply the IQR.
        If capping_method='mad' delta is the value to multiply the MAD.
        If capping_method='quantiles', delta is the percentile on each tail that should be censored. 
            For example, if delta=0.05, the limits will be the 5th and 95th percentiles. 
            If delta=0.1, the limits will be the 10th and 90th percentiles.
        Args:
            variable: str, feature variable name.
            method: must be one of ['norm', 'gaussian', 'iqr', 'mad', 'quantiles']
            delta: float, default 0.9545
            tail: str, default 'right', one of ['left', 'right', 'both'], statistical distribution boundary.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'method':method, 'delta':delta, 'tail':tail,
                           'name':variable if name is None else name},
                  'type':'numerical_outlier', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def numerical_relative(self, variable, reference, function, name=None, keep=True):
        """feature combine transform.

        supported method:
            ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'max', 'min', 'mean']
        
        Args:
            variable: str or str of list, feature variable name.
            reference: str or str of list, feature variable name.
            function: str or str of list, 
                one of ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'max', 'min', 'mean']
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'reference':reference, 'name':'relative' if name is None else name},
                  'type':'numerical_relative', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    
    


from linora.train._visual import *
from linora.train._callback import *
from linora.train._csvlogger import *
from linora.train._earlystopping import *
from linora.train._modelcheckpoint import *
from linora.train._remotemonitor import *
from linora.train._terminateonnan import *
from linora.train import lr


class Csd():
    def add_csd(self, name, xdata, ydata, **kwargs):
        """Plot the cross-spectral density.

        The cross spectral density :math:`P_{xy}` by Welch's average
        periodogram method.  The vectors *x* and *y* are divided into
        *NFFT* length segments.  Each segment is detrended by function
        *detrend* and windowed by function *window*.  *noverlap* gives
        the length of the overlap between segments.  The product of
        the direct FFTs of *x* and *y* are averaged over each segment
        to compute :math:`P_{xy}`, with a scaling to correct for power
        loss due to windowing.

        If len(*x*) < *NFFT* or len(*y*) < *NFFT*, they will be zero
        padded to *NFFT*.

        Args:
            name: data name.
            xdata: x-axis data.
            ydata: y-axis data.
            Fs : float, default: 2
                The sampling frequency (samples per time unit).  It is used to calculate
                the Fourier frequencies, *freqs*, in cycles per time unit.

            window : callable or ndarray, default: `.window_hanning`
                A function or a vector of length *NFFT*.  To create window vectors see
                `.window_hanning`, `.window_none`, `numpy.blackman`, `numpy.hamming`,
                `numpy.bartlett`, `scipy.signal`, `scipy.signal.get_window`, etc.  If a
                function is passed as the argument, it must take a data segment as an
                argument and return the windowed version of the segment.

            sides : {'default', 'onesided', 'twosided'}, optional
                Which sides of the spectrum to return. 'default' is one-sided for real
                data and two-sided for complex data. 'onesided' forces the return of a
                one-sided spectrum, while 'twosided' forces two-sided.

            pad_to : int, optional
                The number of points to which the data segment is padded when performing
                the FFT.  This can be different from *NFFT*, which specifies the number
                of data points used.  While not increasing the actual resolution of the
                spectrum (the minimum distance between resolvable peaks), this can give
                more points in the plot, allowing for more detail. This corresponds to
                the *n* parameter in the call to fft(). The default is None, which sets
                *pad_to* equal to *NFFT*

            NFFT : int, default: 256
                The number of data points used in each block for the FFT.  A power 2 is
                most efficient.  This should *NOT* be used to get zero padding, or the
                scaling of the result will be incorrect; use *pad_to* for this instead.

            detrend : {'none', 'mean', 'linear'} or callable, default: 'none'
                The function applied to each segment before fft-ing, designed to remove
                the mean or linear trend.  Unlike in MATLAB, where the *detrend* parameter
                is a vector, in Matplotlib is it a function.  The :mod:`~matplotlib.mlab`
                module defines `.detrend_none`, `.detrend_mean`, and `.detrend_linear`,
                but you can use a custom function as well.  You can also use a string to
                choose one of the functions: 'none' calls `.detrend_none`. 'mean' calls
                `.detrend_mean`. 'linear' calls `.detrend_linear`.

            scale_by_freq : bool, default: True
                Whether the resulting density values should be scaled by the scaling
                frequency, which gives density in units of Hz^-1.  This allows for
                integration over the returned frequency values.  The default is True for
                MATLAB compatibility.

            noverlap : int, default: 0 (no overlap)
                The number of points of overlap between segments.

            Fc : int, default: 0
                The center frequency of *x*, which offsets the x extents of the
                plot to reflect the frequency range used when a signal is acquired
                and then filtered and downsampled to baseband.

            return_line : bool, default: False
                Whether to include the line object plotted in the returned values.
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]
        elif isinstance(kwargs['color'], dict):
            kwargs['color'] = kwargs.pop('color')['mode']
        kwargs['label'] = name
        
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['ydata'] = ydata
        self._params.ydata[name]['plotmode'] = 'csd'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_csd
        return self
    
    def _execute_plot_csd(self, fig, ax, i, j):
        ax_plot = ax.csd(j['xdata'], j['ydata'], **j['kwargs'])



import numpy as np
from PIL import Image, ImageOps

__all__ = ['add', 'multiply', 'normalize_global', 'normalize_channel', 'normalize_posterize']


def add(image, scale, wise='pixel', prob=1, p=1):
    """add apply to image.
    
    new pixel = int(image + scale)
    Args:
        image: a PIL instance.
        scale: if int or float, value add with image.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying operate.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    if wise=='pixel':
        return image.point(lambda x:x+scale if np.random.uniform()<prob else x)
    elif wise=='channel':
        split = list(image.split())
        for i in range(len(split)):
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:x+scale)
        return Image.merge(image.mode, split)
    elif wise=='pixelchannel':
        img = image.point(lambda x:x+scale)
        axis = [(i,j) for i in range(image.width) for j in range(image.height)]
        np.random.shuffle(axis)
        if prob>0.5:
            axis = axis[:int(len(axis)*(1-prob))]
            for i in axis:
                img.putpixel(i, image.getpixel(i))
            return img
        else:
            axis = axis[:int(len(axis)*prob)]
            for i in axis:
                image.putpixel(i, img.getpixel(i))
            return image
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        for i in wise:
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:x+scale)
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` should be 'pixel' or 'channel' or 'pixelchannel' or list of channel.")


def multiply(image, scale, wise='pixel', prob=1, p=1):
    """Rescale apply to image.
    
    new pixel = int(image * scale)
    Args:
        image: a PIL instance.
        scale: if int or float, value multiply with image.
               if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
        wise: 'pixel' or 'channel' or 'pixelchannel' or list of channel, method of applying operate.
        prob: probability of every pixel or channel being changed.
        p: probability that the image does this. Default value is 1.
    Returns:
        a PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    if wise=='pixel':
        return image.point(lambda x:x*scale if np.random.uniform()<prob else x)
    elif wise=='channel':
        split = list(image.split())
        for i in range(len(split)):
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:x*scale)
        return Image.merge(image.mode, split)
    elif wise=='pixelchannel':
        img = image.point(lambda x:x*scale)
        axis = [(i,j) for i in range(image.width) for j in range(image.height)]
        np.random.shuffle(axis)
        if prob>0.5:
            axis = axis[:int(len(axis)*(1-prob))]
            for i in axis:
                img.putpixel(i, image.getpixel(i))
            return img
        else:
            axis = axis[:int(len(axis)*prob)]
            for i in axis:
                image.putpixel(i, img.getpixel(i))
            return image
    elif isinstance(wise, (list, tuple)):
        split = list(image.split())
        for i in wise:
            if np.random.uniform()<prob:
                split[i] = split[i].point(lambda x:x*scale)
        return Image.merge(image.mode, split)
    else:
        raise ValueError("`wise` should be 'pixel' or 'channel' or 'pixelchannel' or list of channel.")


def normalize_global(image, mean=None, std=None, p=1):
    """Normalize scales `image` to have mean and variance.
    
    This op computes `(x - mean) / std`.
    Args:
        image: a numpy array. shape is `[height, width, channels]` or `[height, width]`.
        mean: if None, computes image mean.
              if int or float, customize image all channels mean.
              if tuple or list, randomly picked in the interval `[mean[0], mean[1])`
        std: if None, computes image std.
             if int or float, customize image all channels std.
             if tuple or list, randomly picked in the interval `[std[0], std[1])`
        p: probability that the image does this. Default value is 1.
    Returns:
        a numpy array.
    """
    if np.random.uniform()>p:
        return image
    if mean is None:
        mean = image.mean()
    elif isinstance(mean, (tuple, list)):
        mean = np.random.uniform(mean[0], mean[1])
    if std is None:
        std = image.std()
    elif isinstance(std, (tuple, list)):
        std = np.random.uniform(std[0], std[1])
    return (image-mean)/std


def normalize_channel(image, mean=None, std=None, p=1):
    """Normalize scales `image` to have mean and variance.
    
    This op computes `(x - mean) / std`.
    Args:
        image: a numpy array. shape is `[height, width, channels]` or `[height, width]`.
        mean: if None, computes image mean.
              if tuple or list, customize image each channels mean, shape should 3 dims.
        std: if None, computes image std.
             if tuple or list, customize image each channels std, shape should 3 dims.
        p: probability that the image does this. Default value is 1.
    Returns:
        a numpy array. The standardized image with same shape as `image`.
    """
    if np.random.uniform()>p:
        return image
    if mean is None:
        mean = [image[:,:,i].mean() for i in range(image.shape[2])]
    elif isinstance(mean, (int, float)):
        mean = [mean for i in range(image.shape[2])]
    if std is None:
        std = [image[:,:,i].std() for i in range(image.shape[2])]
    elif isinstance(std, (int, float)):
        std = [std for i in range(image.shape[2])]
    return (image-mean)/std


def normalize_posterize(image, bits, p=1):
    """Reduce the number of bits for each color channel.
    
    There are up to 2**bits types of pixel values per channel.
    Args:
        image: a PIL instance.
        bits: int or tuple or list, The number of bits to keep for each channel (1-8).
              if list or tuple, randomly picked in the interval `[bits[0], bits[1])` value.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if isinstance(bits, (list, tuple)):
        bits = np.random.randint(bits[0], bits[1])
    return ImageOps.posterize(image, bits)


import itertools
import subprocess

from linora.utils._config import Config

__all__ = ['freeze', 'upgrade', 'upgradeable', 'install', 'uninstall', 
           'mirror', 'download', 'show', 'mirror_set', 'mirror_get',
           'env_view', 'env_clone', 'env_create', 'env_export', 'env_import', 'env_remove']


mirror = Config()
mirror.pip = "https://pypi.org/simple"
mirror.tsinghua = "https://pypi.tuna.tsinghua.edu.cn/simple"
mirror.aliyun = "https://mirrors.aliyun.com/pypi/simple"
mirror.ustc = "https://mirrors.ustc.edu.cn/pypi/web/simple"
mirror.tencent = 'https://mirrors.cloud.tencent.com/pypi/simple'
mirror.douban = 'https://pypi.doubanio.com/simple/'

def libraries_name(name):
    for i in ['==', '>=', '<=', '>', '<']:
        if i in name:
            return name.split(i)+[i]
    for i,j in enumerate(name):
        if j=='-' and name[i+1] in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
            return [name[:i], name[i+1:], '-']
    return [name, '', '']

def download(root, name, mirror=mirror.aliyun, py=''):
    """Download python libraries to the specified folder.
    
    Args:
        root: str, dirs.
        name: str or list. libraries name.
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                mirror=la.utils.pip.mirror.pip
        py: python environment.one of ['', '3'].
    Return:
        root: libraries path list.
    """
    if isinstance(name, str):
        if name.endswith('.txt'):
            cmd = f"pip{py} download -r {name} -d {root} -i {mirror}"
            s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
            s = [i.split(': ') for i in s.decode('utf-8').split('\n')[:-1]]
            s = [i[6:] for i in itertools.chain.from_iterable(s) if 'Saved ' in i]
            return s
        name = [name]
    cmd = f"pip{py} download {' '.join(name)} -d {root} -i {mirror}"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = [i.split(': ') for i in s.decode('utf-8').split('\n')[:-1]]
    s = [i[6:] for i in itertools.chain.from_iterable(s) if 'Saved ' in i]
    return s

def freeze(name=None, py=''):
    """List all python libraries.
    
    Args:
        name: str or list. libraries name.
        py: python environment.one of ['', '3'].
    Return:
        a dict of python libraries version.
    """
    cmd = f"pip{py} list"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = [i.split(' ') for i in s.decode('utf-8').split('\n')[2:-1]]
    s = {i[0]: i[-1] for i in s}
    if not name:
        return s
    if isinstance(name, str):
        name = [name]
    name = [libraries_name(i)[0] for i in name]
    s = {i:s.get(i, '') for i in name}
    return s

def mirror_get(py=''):
    """Get up pip mirrors on your machine.
    
    Args:
        py: python environment.one of ['', '3'].
    Return:
        mirror path.
    """
    cmd = f"pip{py} config list"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    return s.decode('utf-8').strip().split('global.index-url=')[1][1:-1]

def install(name, mirror=mirror.aliyun, py=''):
    """Install python libraries.
    
    Args:
        name: str or list. libraries name. 
              eg. name = 'numpy' or 'numpy==1.0.0' or ['numpy', 'pandas>1.0.0']
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                or mirror=la.utils.pip.mirror.pip
        py: python environment.one of ['', '3'].
    Return:
        a dict of python libraries version.
    """
    if isinstance(name, str):
        if name.startswith('https://github.com/'):
            cmd = f"pip{py} install git+{name}.git"
            s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
            s = s.decode('utf-8').strip().split('\n')[-1].split(' ')[2:]
            if len(s)>1:
                return 'Install Failed'
            return freeze(name=s, py=py)
        elif name.endswith('.whl'):
            cmd = f"pip{py} install {name}"
            s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
            s = s.decode('utf-8').strip().split('\n')[-1].split(' ')[2:]
            return freeze(name=s, py=py)
        elif name.endswith('.txt'):
            cmd = f"pip{py} install -r {name}"
            s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
            s = s.decode('utf-8').strip().split('\n')[-1].split(' ')[2:]
            return freeze(name=s, py=py)
        name = [name]
    name = [libraries_name(i) for i in name]
    name1 = [i[0] for i in name]
    name2 = [i[0] for i in name if i[2]=='']
    name2 = name2 + [i[0]+'=='+i[1] for i in name if i[2]=='-']
    name2 = name2 + [i[0]+i[2]+i[1] for i in name if i[2] not in ['', '-']]
    name2 = ['"'+i+'"' for i in name2]
    cmd = f"pip{py} install {' '.join(name2)} -i {mirror}"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    return freeze(name=name1, py=py)

def search(name, mirror=mirror.aliyun, py=''):
    """Show python libraries.
    
    Args:
        name: str. libraries name.
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                or mirror=la.utils.pip.mirror.pip
        py: python environment.one of ['', '3'].
    Return:
        a dict of python libraries version information.
    """
    assert isinstance(name, str), "`name` should be str."
    cmd = f"pip{py} install pip-search -U -i {mirror}"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    cmd = f"pip_search{py} {name}"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = [i.strip() for i in s.decode('utf-8').split('\n')[:-1]]
    s = [s[r:-1] for r, i in enumerate(s) if 'numpy' in i][1]
    s = [[i.strip() for i in j[3:-1].split('│')] for j in s]
    result = {}
    for i in s:
        if i[0]!='':
            result[i[0]] = {'version':i[1], 'released':i[2], 'description':i[3]}
            t = i[0]
        else:
            result[t]['description'] += ' '+i[3]
    return result

def mirror_set(mirror, py=''):
    """Set up pip mirrors on your machine.
    
    Args:
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                or mirror=la.utils.pip.mirror.pip
        py: python environment.one of ['', '3'].
    Return:
        mirror file path.
    """
    subprocess.Popen(f"pip{py} install pip -U -i {mirror}", stdout=subprocess.PIPE, shell=True).communicate()[0]
    cmd = f"pip{py} config set global.index-url {mirror}"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    return s.decode('utf-8').strip()

def show(name, py=''):
    """Show python libraries.
    
    Args:
        name: str or list. libraries name.
        py: python environment.one of ['', '3'].
    Return:
        a dict of python libraries version information.
    """
    if isinstance(name, str):
        name = [name]
    name = [libraries_name(i)[0] for i in name]
    name = freeze(name, py)
    t = {}
    for lib in name:
        if name[lib]=='':
            t[lib] = {}
        else:
            cmd = f"pip{py} show {lib}"
            s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
            s = [i.split(': ') for i in s.decode('utf-8').split('\n')[:-1]]
            s = {i[0]: i[1] for i in s}
            t[lib] = s
    return t

def uninstall(name, py=''):
    """Uinstall python libraries.
    
    Args:
        name: str or list. libraries name.
        py: python environment.one of ['', '3'].
    Return:
        uninstall log dict.
    """
    if isinstance(name, str):
        name = [name]
    name = [libraries_name(i)[0] for i in name]
    name = freeze(name, py)
    name1 = {i:'Notexist' for i in name if name[i]==''}
    name = [i for i in name if name[i]!='']
    cmd = f"pip{py} uninstall {' '.join(name)} -y"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = ' '.join([i.strip() for i in s.decode('utf-8').strip().split('\n') if 'Successfully uninstalled' in i])
    s = {i:'Success' if i in s else 'Failed' for i in name}
    return {**s, **name1}

def upgrade(name, version=None, mirror=mirror.aliyun, py='', logger=False):
    """Upgrade python libraries.
    
    Args:
        name: str or list. libraries name.
        version: str or list. libraries version.
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set eg. mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                or eg. mirror=la.utils.pip.mirror.tsinghua
        py: python environment.one of ['', '3'].
        logger: whether to print the log.
    Return:
        a dict of python libraries version.
    """
    assert version is None or isinstance(version, (str, list)), "`version` should be None or str or list."
    if isinstance(name, str):
        name = [name]
    old_lib = freeze(name=name, py=py)
    if version is not None:
        if isinstance(version, str):
            version = [version]
        assert len(name)==len(version), "`name` and `version` should be same number."
        for (dist, ver) in zip(name, version):
            if ver=='':
                cmd = f"pip{py} install --upgrade {dist} -i {mirror}"
            else:
                if len([i for i in ['==', '>', '<', '>=', '<='] if i in ver])==0:
                    ver = '=='+ver
                cmd = f"pip{py} install --upgrade {dist}{ver} -i {mirror}"
            if logger:
                subprocess.call(cmd, shell=True)
            else:
                subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    else:
        for dist in old_lib:
            cmd = f"pip{py} install --upgrade {dist} -i {mirror}"
            if logger:
                subprocess.call(cmd, shell=True)
            else:
                subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    new_lib = freeze(name=name, py=py)
    lib = {i:{'old_version': old_lib[i], "new_version":new_lib[i]} for i in new_lib}
    return lib

def upgradeable(mirror=mirror.aliyun, py=''):
    """Veiw upgradeable python libraries.
    
    Args:
        mirror: pip install libraries mirror,
                default official https://pypi.org/simple.
                or you can set mirror='https://pypi.tuna.tsinghua.edu.cn/simple'.
                or mirror=la.utils.pip.mirror.pip
        py: python environment.one of ['', '3'].
    Return:
        a dict of python libraries version.
    """
    cmd = f"pip{py} list -o -i {mirror}"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = [i.strip().split(' ') for i in s.decode('utf-8').split('\n')[2:-1]]
    s = [list(filter(lambda x:len(x)>0, i)) for i in s]
    s = {i[0]:{'version':i[1], 'latest':i[2], 'type':i[-1]} for i in s}
    return s

def env_view():
    """Get virtual environment info."""
    cmd = f"conda info -e"
    s = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = s.decode('utf-8').strip().split('\n')[2:]
    s = [i.split(' ') for i in s]
    return {i[0]:i[-1] for i in s}

def env_clone(new_name, old_name):
    """copy already exists virtual environment.
    
    Args:
        new_name: new virtual environment.
        old_name: already exists virtual environment.
    Return:
        log info.
    """
    cmd = 'conda update -n base -c defaults conda'
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    if old_name not in s:
        return f'Virtual environment {old_name} not exists.'
    cmd = f"conda create -n {new_name} --clone {old_name}"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    if new_name in s:
        return 'Virtual environment successfully created.'
    return 'Virtual environment failed created.'

def env_create(name, version):
    """Create virtual environment.
    
    Args:
        name: virtual environment.
        version: python version.
    Return:
        log info.
    """
    cmd = 'conda update -n base -c defaults conda'
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    if name in s:
        return 'Virtual environment already exists.'
    cmd = f"conda create -n {name} python={version} -y"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    if name in s:
        return 'Virtual environment successfully created.'
    return 'Virtual environment failed created.'

def env_export(yml):
    """Export virtual environment to `.yml` format file.
    
    Args:
        yml: `.yml` format file.
    Return:
        log info.
    """
    cmd = 'conda update -n base -c defaults conda'
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    cmd = f"conda env export > {yml}"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    return yml

def env_import(yml):
    """Import virtual environment from `.yml` format file.
    
    Args:
        yml: `.yml` format file.
    Return:
        log info.
    """
    cmd = 'conda update -n base -c defaults conda'
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    cmd = f"conda env create -f {yml}"
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s1 = view_env()
    if len(s1)>len(s):
        return 'Virtual environment successfully created.'
    return 'Virtual environment failed created.'

def env_remove(name):
    """Remove virtual environment.
    
    Args:
        name: virtual environment.
    Return:
        log info.
    """
    s = view_env()
    if name not in s:
        return 'Virtual environment not exists.'
    cmd = f'conda remove -n {name} --all'
    subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True).communicate()[0]
    s = view_env()
    if name not in s:
        return 'Virtual environment successfully removed.'
    return 'Virtual environment failed removed.'


from linora.utils._config import *
from linora.utils._email import *
from linora.utils._logger import *
from linora.utils._progbar import *
from linora.utils._schedulers import *
from linora.utils import message
from linora.utils import pip



import io
from setuptools import setup, find_packages

def readme():
    with io.open('README.md', encoding='utf-8') as f:
        return f.read()

setup(name='linora',
      version='2.0.0rc1',
      install_requires=[
          'pandas>=1.5.2', 
          'Pillow>=9.3.0',
          'joblib>=1.2.0',
          'requests>=2.28.0',
          'rarfile',
          'av'
      ],
      description='Simple and efficient tools for data mining and data analysis.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/Hourout/linora',
      author='JinQing Lee',
      author_email='hourout@163.com',
      keywords=['hyperparameter-optimization', 'XGBoost', 'LightGBM', 'data-mining', 
                'data-analysis', 'machine-learning', 'image', 'text', 'data-science', 
                'logging', 'parallel', 'feature-engineering', 'metrics', 'schedulers', 'datasets'],
      license='Apache License Version 2.0',
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Scientific/Engineering :: Visualization',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11'
      ],
      packages=find_packages(),
      zip_safe=False)



class Pie():
    def add_pie(self, name, xdata, **kwargs):
        """Make a pie plot.
        
        Args:
            name: data name.
            xdata: The wedge sizes.
            explode : array-like, default: None
                If not *None*, is a ``len(x)`` array which specifies the fraction
                of the radius with which to offset each wedge.

            labels : list, default: None
                A sequence of strings providing the labels for each wedge

            colors : array-like, default: None
                A sequence of colors through which the pie chart will cycle.  If
                *None*, will use the colors in the currently active cycle.

            autopct : None or str or callable, default: None
                If not *None*, is a string or function used to label the wedges
                with their numeric value.  The label will be placed inside the
                wedge.  If it is a format string, the label will be ``fmt % pct``.
                If it is a function, it will be called.

            pctdistance : float, default: 0.6
                The ratio between the center of each pie slice and the start of
                the text generated by *autopct*.  Ignored if *autopct* is *None*.

            shadow : bool, default: False
                Draw a shadow beneath the pie.

            normalize : None or bool, default: None
                When *True*, always make a full pie by normalizing x so that
                ``sum(x) == 1``. *False* makes a partial pie if ``sum(x) <= 1``
                and raises a `ValueError` for ``sum(x) > 1``.

                When *None*, defaults to *True* if ``sum(x) >= 1`` and *False* if
                ``sum(x) < 1``.

                Please note that the previous default value of *None* is now
                deprecated, and the default will change to *True* in the next
                release. Please pass ``normalize=False`` explicitly if you want to
                draw a partial pie.

            labeldistance : float or None, default: 1.1
                The radial distance at which the pie labels are drawn.
                If set to ``None``, label are not drawn, but are stored for use in
                ``legend()``

            startangle : float, default: 0 degrees
                The angle by which the start of the pie is rotated,
                counterclockwise from the x-axis.

            radius : float, default: 1
                The radius of the pie.

            counterclock : bool, default: True
                Specify fractions direction, clockwise or counterclockwise.

            wedgeprops : dict, default: None
                Dict of arguments passed to the wedge objects making the pie.
                For example, you can pass in ``wedgeprops = {'linewidth': 3}``
                to set the width of the wedge border lines equal to 3.
                For more details, look at the doc/arguments of the wedge object.
                By default ``clip_on=False``.

            textprops : dict, default: None
                Dict of arguments to pass to the text objects.

            center : (float, float), default: (0, 0)
                The coordinates of the center of the chart.

            frame : bool, default: False
                Plot Axes frame with the chart if true.

            rotatelabels : bool, default: False
                Rotate each label to the angle of the corresponding slice if true.
        """
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['plotmode'] = 'pie'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_pie
        return self
    
    def _execute_plot_pie(self, fig, ax, i, j):
        axplot = ax.pie(j['xdata'], **j['kwargs'])


import numpy as np
from PIL import ImageFilter

__all__ = ['FilterMode', 'filters', 'convolution']


def convolution(image, kernel, scale=None, offset=0):
    """Create a convolution kernel.  
    
    The current version only supports 3x3 and 5x5 integer and floating point kernels.
    In the current version, kernels can only be applied to "L" and "RGB" images.

    Args:
        image: a PIL instance.
        kernel: a sequence containing kernel weights.
        scale: each pixel is divided by this value, the default is the sum of the kernel weights.
        offset: this value is added to the result, after it has been divided by the scale factor.
    Return:
        a PIL instance.
    """
    kernel = np.array(kernel).flatten().tolist()
    if len(kernel)==9:
        size = (3,3)
    elif len(kernel)==25:
        size = (5,5)
    else:
        raise ValueError('Kernel size must be (3,3) or (5,5).')
    return image.filter(ImageFilter.Kernel(size, kernel, scale, offset))


class filter_mode:
    """image blur type."""
    Box               = 'box'
    Gaussian          = 'gaussian'
    Unsharpmask       = 'unsharpmask'
    Rank              = 'rank'
    Median            = 'median'
    Min               = 'min'
    Max               = 'max'
    Mode              = 'mode'
    Mean              = 'BLUR'
    CONTOUR           = 'CONTOUR'
    DETAIL            = 'DETAIL'
    EDGE_ENHANCE      = 'EDGE_ENHANCE'
    EDGE_ENHANCE_MORE = 'EDGE_ENHANCE_MORE'
    EMBOSS            = 'EMBOSS'
    FIND_EDGES        = 'FIND_EDGES'
    SHARPEN           = 'SHARPEN'
    SMOOTH            = 'SMOOTH'
    SMOOTH_MORE       = 'SMOOTH_MORE'
    
FilterMode = filter_mode()


def filters(image, mode=FilterMode.Mean, p=1, **kwarg):
    """
    la.image.FilterMode.Box:
        Blurs the image by setting each pixel to the average value of the pixels 
        in a square box extending radius pixels in each direction. 
        Supports float radius of arbitrary size. Uses an optimized implementation 
        which runs in linear time relative to the size of the image for any radius value.
        
        you should append param `radius`
        radius: Size of the box in one direction. Radius 0 does not blur, returns an identical image. 
                Radius 1 takes 1 pixel in each direction, i.e. 9 pixels in total.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Box, radius=2)
    
    la.image.FilterMode.Gaussian:
        Gaussian blur filter.
        
        you should append param `radius`
        radius: Blur radius.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Gaussian, radius=2)
        
    la.image.FilterMode.Unsharpmask:
        Unsharp mask filter.
        See Wikipedia’s entry on digital unsharp masking for an explanation of the parameters.
        
        you should append param `radius`,`percent`,`threshold`
        radius: Blur radius.
        percent: Unsharp strength, in percent
        threshold: Threshold controls the minimum brightness change that will be sharpened
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Unsharpmask, radius=2, percent=150, threshold=3)
        
    la.image.FilterMode.Rank:
        Create a rank filter. The rank filter sorts all pixels in a window of the given size, and returns the rank’th value.
        
        you should append param `size`,`rank`
        size: The kernel size, in pixels.
        rank: What pixel value to pick. Use 0 for a min filter, 
              size * size / 2 for a median filter, size * size - 1 for a max filter, etc.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Rank, size, rank)
        
    la.image.FilterMode.Median:
        Create a median filter. Picks the median pixel value in a window with the given size.
        
        you should append param `size`
        size: The kernel size, in pixels.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Median, size=3)
        
    la.image.FilterMode.Min:
        Create a min filter. Picks the lowest pixel value in a window with the given size.
        
        you should append param `size`
        size: The kernel size, in pixels.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Min, size=3)
        
    la.image.FilterMode.Max:
        Create a max filter. Picks the largest pixel value in a window with the given size.
        
        you should append param `size`
        size: The kernel size, in pixels.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Max, size=3)
        
    la.image.FilterMode.Mode:
        Create a mode filter. Picks the most frequent pixel value in a box with the given size. 
        
        Pixel values that occur only once or twice are ignored; 
        if no pixel value occurs more than twice, the original pixel value is preserved.
        
        you should append param `size`
        size: The kernel size, in pixels.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Mode, size=3)
        
    la.image.FilterMode.Mean:
        Normal blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.Mean)
        
    la.image.FilterMode.CONTOUR:
        contour blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.CONTOUR)
        
    la.image.FilterMode.DETAIL:
        detail blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.DETAIL)
        
    la.image.FilterMode.EDGE_ENHANCE:
        Edge enhancement blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.EDGE_ENHANCE)
        
    la.image.FilterMode.EDGE_ENHANCE_MORE:
        Edge enhancement more blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.EDGE_ENHANCE_MORE)
        
    la.image.FilterMode.EMBOSS:
        emboss blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.EMBOSS)
        
    la.image.FilterMode.FIND_EDGES:
        Find the edge blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.FIND_EDGES)
        
    la.image.FilterMode.SHARPEN:
        Sharpen blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.SHARPEN)
    
    la.image.FilterMode.SMOOTH:
        Smooth blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.SMOOTH)
        
    la.image.FilterMode.SMOOTH_MORE:
        Smooth threshold blur.
        eg.
        la.image.filters(image, mode=la.image.FilterMode.SMOOTH_MORE)

    Args:
        image: a PIL instance.
        mode:la.image.FilterMode.
        p: probability that the image does this. Default value is 1.
    Returns:
        A PIL instance.
    """
    if np.random.uniform()>p:
        return image
    if 'radius' not in kwarg:
        kwarg['radius'] = 2
    if 'size' not in kwarg:
        kwarg['size'] = 3
    if 'percent' not in kwarg:
        kwarg['percent'] = 150
    if 'threshold' not in kwarg:
        kwarg['threshold'] = 3
    if 'rank' not in kwarg:
        kwarg['rank'] = 1
    if mode=='box':
        return image.filter(ImageFilter.BoxBlur(radius=kwarg['radius']))
    elif mode=='gaussian':
        return image.filter(ImageFilter.GaussianBlur(radius=kwarg['radius']))
    elif mode=='unsharpmask':
        return image.filter(ImageFilter.UnsharpMask(radius=kwarg['radius'], percent=kwarg['percent'], threshold=kwarg['threshold']))
    elif mode=='rank':
        return image.filter(ImageFilter.RankFilter(size=kwarg['size'], rank=kwarg['rank']))
    elif mode=='median':
        return image.filter(ImageFilter.MedianFilter(size=kwarg['size']))
    elif mode=='min':
        return image.filter(ImageFilter.MinFilter(size=kwarg['size']))
    elif mode=='max':
        return image.filter(ImageFilter.MaxFilter(size=kwarg['size']))
    elif mode=='mode':
        return image.filter(ImageFilter.ModeFilter(size=kwarg['size']))
    elif mode=='BLUR':
        return image.filter(ImageFilter.BLUR)
    elif mode=='CONTOUR':
        return image.filter(ImageFilter.CONTOUR)
    elif mode=='DETAIL':
        return image.filter(ImageFilter.DETAIL)
    elif mode=='EDGE_ENHANCE':
        return image.filter(ImageFilter.EDGE_ENHANCE)
    elif mode=='EDGE_ENHANCE_MORE':
        return image.filter(ImageFilter.EDGE_ENHANCE_MORE)
    elif mode=='EMBOSS':
        return image.filter(ImageFilter.EMBOSS)
    elif mode=='FIND_EDGES':
        return image.filter(ImageFilter.FIND_EDGES)
    elif mode=='SHARPEN':
        return image.filter(ImageFilter.SHARPEN)
    elif mode=='SMOOTH':
        return image.filter(ImageFilter.SMOOTH)
    elif mode=='SMOOTH_MORE':
        return image.filter(ImageFilter.SMOOTH_MORE)
    else:
        raise ValueError("mode must be la.image.FilterMode param")




# from linora.chart._metrics import *
# from linora.chart._feature import *

from linora.chart._config import *
from linora.chart._grid import *
from linora.chart._plot import *


# from linora.chart._errorbar import *
# from linora.chart._fillline import *
# from linora.chart._hist import *
# from linora.chart._hist2d import *
# from linora.chart._line import *
# from linora.chart._scatter import *



# import av
# import numpy as np

from linora.image._image_util import list_images
# from linora.vedio._vedio_io import Vedio

__all__ = ['list_vedios']


def list_vedios(directory, file_format=('mpeg', 'mpg', 'dat', 'mp4', 'avi', 'mov', 'asf', 'wmv', 'mkv', 'flv', 'rmvb')):
    """Lists all vedios in a directory, including all subdirectories.
    
    Args:
        directory: string, absolute path to the directory.
        file_format: tuple of strings or single string, extensions of the vedios.
    Returns:
        a list of vedio paths.
    """
    return list_images(directory, file_format=file_format)


# def vedio_to_stream(vedio, batch=1, data_format='HWCN', dtype='float32'):
#     """Converts a Vedio instance to a Numpy array.
    
#     Args:
#         image: Vedio instance.
#         batch: each iter batch number.
#         data_format: array data format, eg.'HWCN', 'CHWN'. 'image' return pillow instance.
#         dtype: Dtype to use for the returned array.
#     Returns:
#         A Numpy array iterator.
#     """
#     container = av.open(vedio._filename)
#     container.streams.video[0].thread_type = "AUTO"
#     stream = container.streams.video[0]
#     batch_array = []
    
#     if data_format=='image':
#         for r, frame in enumerate(container.decode(stream), start=1):
#             batch_array.append(frame.to_image())
#             if not r%batch:
#                 yield batch_array
#                 batch_array = []
#         if batch_array:
#             yield batch_array
#     else:
#         transpose = {'H':0, 'W':1, 'C':2, 'N':3}
#         for r, frame in enumerate(container.decode(stream), start=1):
#             batch_array.append(np.expand_dims(frame.to_ndarray(format="rgb24"), axis=-1))
#             if not r%batch:
#                 array = np.concatenate(batch_array, axis=-1, dtype=dtype).transpose(tuple(transpose[i] for i in data_format))
#                 batch_array = []
#                 yield array
#         if batch_array:
#             yield np.concatenate(batch_array, axis=-1, dtype=dtype).transpose(tuple(transpose[i] for i in data_format))
#     container.close()

    
# def vedio_to_array(vedio, data_format='HWCN', dtype='float32'):
#     """Converts a Vedio instance to a 4D Numpy array.
    
#     Args:
#         image: Vedio instance.
#         data_format: array data format, eg.'HWCN', 'CHWN'.
#         dtype: Dtype to use for the returned array.
#     Returns:
#         A 4D Numpy array.
#     """
#     container = av.open(vedio._filename)
#     container.streams.video[0].thread_type = "AUTO"
#     stream = container.streams.video[0]
#     transpose = {'H':0, 'W':1, 'C':2, 'N':3}
#     data = np.concatenate([np.expand_dims(frame.to_ndarray(format="rgb24"), axis=-1) for frame in container.decode(stream)],
#                           axis=-1, dtype=dtype)
#     if data_format!='HWCN':
#         data = data.transpose(tuple(transpose[i] for i in data_format))
#     container.close()
#     return data


# def array_to_vedio(x, data_format='HWCN', fps=30):
#     """Converts a 4D Numpy array to a Vedio instance.
    
#     Args:
#         x: a 4D Numpy array.
#         data_format: array data format, eg.'HWCN'.
#         fps: Frames per second.
#     Returns:
#         Vedio instance.
#     """
#     vedio_params = {'vedio_duration':x.shape[data_format.find('N')]/fps, 
#                     'vedio_shape':(x.shape[data_format.find('W')], x.shape[data_format.find('H')]),
#                     'vedio_fps':fps,
#                     'vedio_frames':x.shape[data_format.find('N')]
#                     }
#     transpose = {i:r for r,i in enumerate(data_format)}
#     vedio = Vedio(params=vedio_params, data=x.transpose(tuple(transpose[i] for i in 'HWCN')).astype('uint8'), fps=fps)
#     return vedio


import math

import numpy as np
import pandas as pd

from linora.metrics._utils import _sample_weight

__all__ = ['euclidean', 'manhattan', 'chebyshev', 'minkowski', 'hamming', 'jaccard', 'pearson', 'cosine', 
           'levenshtein', 'dice', 'ochiia', 'braycurtis', 'geodesic', 'canberra', 'hausdorff', 'chisquare', 
           'hellinger', 'bhattacharyya', 'wasserstein'
          ]


def euclidean(x, y, normalize=False, sample_weight=None):
    """euclidean distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        normalize: default False, std=pd.concat([x, y]).std() if normalize else 1.
        sample_weight: list or array of sample weight.
    Returns:
        euclidean distance value.
    """
    x = pd.Series(x)
    y = pd.Series(y)
    sample_weight = _sample_weight(x, sample_weight)
    std = pd.concat([x, y], axis=1).std(axis=1) if normalize else 1
    return np.sqrt(np.sum(np.square((x-y)*sample_weight/std)))


def manhattan(x, y, sample_weight=None):
    """manhattan distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        manhattan distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return np.sum(np.abs(np.array(x)-np.array(y))*sample_weight)


def chebyshev(x, y, sample_weight=None):
    """chebyshev distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        chebyshev distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return np.max((np.array(x)-np.array(y))*sample_weight)


def minkowski(x, y, p, sample_weight=None):
    """minkowski distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        p: int, norm dimension.
        sample_weight: list or array of sample weight.
    Returns:
        minkowski distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return np.power(np.sum(np.power(np.abs(np.array(x)-np.array(y))*sample_weight, p)), 1/p)


def hamming(x, y, sample_weight=None):
    """hamming distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        hamming distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return sum(map(lambda x,y,z:(x!=y)*z, x, y, sample_weight))/sample_weight.sum()


def jaccard(x, y):
    """jaccard distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
    Returns:
        jaccard distance value.
    """
    return 1 - len(set(x).intersection(set(y)))/len(set(x).union(set(y)))


def pearson(x, y, sample_weight=None):
    """pearson distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        pearson distance value.
    """
    x = np.array(x)
    y = np.array(y)
    sample_weight = _sample_weight(x, sample_weight)
    x_mean, y_mean = np.mean(x), np.mean(y)
    d = np.sum((x-x_mean)*(y-y_mean)*sample_weight)
    d2 = np.sqrt(np.sum(np.square(x-x_mean)))*np.sqrt(np.sum(np.square(y-y_mean)))
    return 1-d/d2


def cosine(x, y, sample_weight=None):
    """cosine distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        cosine distance value.
    """
    x = np.array(x)
    y = np.array(y)
    sample_weight = _sample_weight(x, sample_weight)
    return (x*y*sample_weight).sum()/np.sqrt(np.square(x).sum())/np.sqrt(np.square(y).sum())


def levenshtein(x, y, normalize=False, sample_weight=None):
    """levenshtein distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        normalize: bool, default False, if normalize is True, levenshtein distance should be distance/max(len(x), len(y)).
        sample_weight: list or array of sample weight.
    Returns:
        levenshtein distance value.
    """
    def levenshtein1(s1, s2, weight, normalize=False):
        if len(s1) < len(s2):
            return levenshtein1(s2, s1, normalize)
        if not s2:
            return len(s1)

        a = range(len(s2) + 1)
        for i, c1 in enumerate(s1):
            b = [i+1]
            for j, c2 in enumerate(s2):
                b.append(min(a[j+1]+1, b[j]+1, a[j]+(c1 != c2)))
            a = b

        if normalize:
            return (b[-1] / len(s1))*weight
        return b[-1]*weight
    sample_weight = _sample_weight(x, sample_weight)
    return pd.DataFrame({'label1':x, 'label2':y, 'weight':sample_weight}).apply(lambda x:levenshtein1(x[0], x[1], x[2], normalize=normalize), axis=1).sum()


def canberra(x, y, sample_weight=None):
    """canberra distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        canberra distance value.
    """
    assert len(x)==len(y), 'x shape should be same with y.'
    x = pd.Series(x)
    y = pd.Series(y)
    sample_weight = _sample_weight(x, sample_weight)
    return ((x-y)*sample_weight/(x.abs()+y.abs())).sum()


def geodesic(x, y):
    """Calculation of latitude and longitude distance by haversine formula.
    
    Args:
        x: list or tuple, Longitude and latitude coordinates, [latitude, longitude]
        y: list or tuple, Longitude and latitude coordinates, [latitude, longitude]
    Return:
        Longitude and latitude distance, unit is 'km'.
    """
    lat0 = math.radians(x[0])
    lat1 = math.radians(y[0])
    lng0 = math.radians(x[1])
    lng1 = math.radians(y[1])
    h = math.sin(math.fabs(lat0 - lat1)/2)**2 + math.cos(lat0) * math.cos(lat1) * math.sin(math.fabs(lng0 - lng1)/2)**2
    return 2 * 6371.393 * math.asin(math.sqrt(h))


def braycurtis(x, y, sample_weight=None):
    """braycurtis distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        braycurtis distance value.
    """
    x = np.array(x)
    y = np.array(y)
    sample_weight = _sample_weight(x, sample_weight)
    return np.sum(np.abs(x-y)*sample_weight)/(np.sum(x)+np.sum(y))


def ochiia(x, y):
    """ochiia distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
    Returns:
        ochiia distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return 1 - len(set(x).intersection(set(y)))/np.sqrt(len(set(x))*len(set(y)))


def dice(x, y):
    """dice distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
    Returns:
        dice distance value.
    """
    return 1 - 2*len(set(x).intersection(set(y)))/(len(set(x))+len(set(y)))


def hausdorff(x, y, method=euclidean, sample_weight=None):
    """canberra distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        method: distance method, default la.metrics.distance.euclidean, 
                see la.metrics.distance, Some methods are effective.
        sample_weight: list or array of sample weight.
    Returns:
        canberra distance value.
    """
    x = np.array(x)
    y = np.array(y)
    sample_weight = _sample_weight(x, sample_weight)
    fhd = max([min([method(i, j, sample_weight=sample_weight) for j in y]) for i in x])
    rhd = max([min([method(i, j, sample_weight=sample_weight) for j in x]) for i in y])
    return max(fhd,rhd)


def chisquare(x, y, sample_weight=None):
    """chi square distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        chi square distance value.
    """
    x = np.asarray(x, np.int32)
    y = np.asarray(y, np.int32)
    sample_weight = _sample_weight(x, sample_weight)
    value = np.square((x-y)*sample_weight)
    y[np.where(y==0)] = 1
    return np.sum(value/y)


def hellinger(x, y, sample_weight=None):
    """hellinger distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        hellinger distance value.
    """
    sample_weight = _sample_weight(x, sample_weight)
    return 1/np.sqrt(2)*np.linalg.norm((np.sqrt(x)-np.sqrt(y))*sample_weight)


def bhattacharyya(x, y, sample_weight=None):
    """bhattacharyya distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
        sample_weight: list or array of sample weight.
    Returns:
        bhattacharyya distance value.
    """
    x = np.array(x)
    y = np.array(y)
    sample_weight = _sample_weight(x, sample_weight)
    return np.log(np.sum(np.sqrt(x * y*sample_weight)))


def wasserstein(x, y):
    """wasserstein distance.
    
    Args:
        x: pd.Series or array or list, sample n dim feature value.
        y: pd.Series or array or list, sample n dim feature value.
    Returns:
        wasserstein distance value.
    """
    x = np.array(x)
    y = np.array(y)

    x_sort = np.argsort(x)
    y_sort = np.argsort(y)

    z = np.concatenate([x, y])
    z.sort(kind='mergesort')

    x_cdf = x[x_sort].searchsorted(z[:-1], 'right')/x.size
    y_cdf = y[y_sort].searchsorted(z[:-1], 'right')/y.size
    return np.sum(np.multiply(np.abs(x_cdf - y_cdf), np.diff(z)))


from linora.image._image_rescale import *

class ImageRescaleAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def add(self, scale, wise='pixel', prob=1, p=None):
        """add apply to image.

        new pixel = int(image + scale)
        Args:
            scale: if int or float, value add with image.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            wise: 'pixel' or 'channel' or list of channel, method of applying operate.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = add(self.image, scale, wise, prob, p)
        return self
    
    def multiply(self, scale, wise='pixel', prob=1, p=None):
        """Rescale apply to image.

        new pixel = int(image * scale)
        Args:
            scale: if int or float, value multiply with image.
                   if tuple or list, randomly picked in the interval `[scale[0], scale[1])`.
            wise: 'pixel' or 'channel' or list of channel, method of applying operate.
            prob: probability of every pixel or channel being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = multiply(self.image, scale, wise, prob, p)
        return self
        
    def normalize_global(self, mean=None, std=None, p=None):
        """Normalize scales `image` to have mean and variance.

        This op computes `(x - mean) / std`.
        Args:
            mean: if None, computes image mean.
                  if int or float, customize image all channels mean.
                  if tuple or list, randomly picked in the interval `[mean[0], mean[1])`
            std: if None, computes image std.
                 if int or float, customize image all channels std.
                 if tuple or list, randomly picked in the interval `[std[0], std[1])`
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = normalize_global(self.image, mean, std, p)
        return self
    
    def normalize_channel(self, mean=None, std=None, p=None):
        """Normalize scales `image` to have mean and variance.

        This op computes `(x - mean) / std`.
        Args:
            mean: if None, computes image mean.
                  if tuple or list, customize image each channels mean, shape should 3 dims.
            std: if None, computes image std.
                 if tuple or list, customize image each channels std, shape should 3 dims.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = normalize_channel(self.image, mean, std, p)
        return self
    
    def normalize_posterize(self, bits, p=None):
        """Reduce the number of bits for each color channel.

        There are up to 2**bits types of pixel values per channel.
        Args:
            bits: int or tuple or list, The number of bits to keep for each channel (1-8).
                  if list or tuple, randomly picked in the interval `[bits[0], bits[1])` value.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = normalize_posterize(self.image, bits, p)
        return self



import io
import base64

import requests
from PIL import Image, ImageOps

from linora.gfile._gfile import exists

__all__ = ['read_image', 'save_image', 'encode_base64', 'decode_base64']


def read_image(filename):
    """Reads the contents of file to a PIL Image instance.
    
    Args:
        filename: str, image absolute path, or io.BytesIO stream to image file.
    Returns:
        a PIL instance.
    """
    if isinstance(filename, str):
        if exists(filename):
            with open(filename, 'rb') as f:
                image = Image.open(io.BytesIO(f.read()))
        else:
            image = Image.open(io.BytesIO(requests.get(filename).content), 'r')
    elif isinstance(filename, io.BytesIO):
        image = Image.open(filename)
    else:
        raise TypeError('path should be absolute path or io.BytesIO')
    return ImageOps.exif_transpose(image)


def save_image(filename, image, file_format=None, **kwargs):
    """Saves an image stored as a Numpy array to a path or file object.
    
    if save gif image, please use save_image(filename, image, file_format=None, save_all=True, append_images=[im1, im2, ...])
    
    Args
        filename: Path or file object.
        image: A PIL instance.
        file_format: Optional file format override. If omitted, the
            format to use is determined from the filename extension.
            If a file object was used instead of a filename, this
            parameter should always be used.
        **kwargs: Additional keyword arguments passed to `PIL.Image.save()`.
        if save gif, param `duration` and `loop` is optional.
    """
    if file_format is None:
        file_format = filename.split('.')[-1]
    if image.mode == 'RGBA' and file_format in ['jpg', 'jpeg']:
        image = image.convert('RGB')
    if filename.lower().endswith('.gif'):
        assert isinstance(image, list), '`image` must be image of list.'
        duration = kwargs['duration'] if 'duration' in kwargs else 100
        loop = kwargs['loop'] if 'loop' in kwargs else 0
        image[0].save(filename, format='GIF', append_images=image[1:], save_all=True, duration=duration, loop=loop)
    else:
        image.save(filename, format=file_format, **kwargs)


def encode_base64(filename):
    """encode image to string.
    
    Args
        filename: image file path.
    Returns:
        a bites string.
    """
    with open(filename, "rb")as f:
        bs64 = base64.b64encode(f.read()).decode()
    return bs64


def decode_base64(filename, image_str):
    """decode image to file.
    
    Args
        filename: image file path.
        image_str: image bites string.
    """
    with open(filename, "wb") as f:
        f.write(base64.b64decode(image_str))
